#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np
from geopack import geopack
from matplotlib.patches import Wedge

# Local imports
from ..pyrf import datetime642unix, iso86012datetime64, magnetosphere

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _add_earth(ax=None, **kwargs):
    theta1, theta2 = 90.0, 270.0
    nightside_ = Wedge(
        (0.0, 0.0),
        1.0,
        theta1,
        theta2,
        fc="k",
        ec="k",
        **kwargs,
    )
    dayside_ = Wedge(
        (0.0, 0.0),
        1.0,
        theta2,
        theta1,
        fc="w",
        ec="k",
        **kwargs,
    )
    for wedge in [nightside_, dayside_]:
        ax.add_artist(wedge)

    return [nightside_, dayside_]


def _add_field_lines(ax, tint):
    # Get dipole axis at begin of the time interval
    ut = datetime642unix(iso86012datetime64(np.array(tint)))[0]
    _ = geopack.recalc(ut)

    x_lines_m, z_lines_m = [[], []]
    x_lines_p, z_lines_p = [[], []]

    xx_gsm, zz_gsm = np.meshgrid(
        np.linspace(-30, 6, 19),
        np.linspace(-5, 5, 10),
    )
    xx_gsm = np.reshape(xx_gsm, (xx_gsm.size,))
    zz_gsm = np.reshape(zz_gsm, (zz_gsm.size,))

    for x_gsm, z_gsm in zip(xx_gsm, zz_gsm):
        y_gsm = 0
        _, _, _, xx, _, zz = geopack.trace(
            x_gsm,
            y_gsm,
            z_gsm,
            dir=-1,
            rlim=100,
            r0=0.99999,
            parmod=2,
            exname="t89",
            inname="igrf",
            maxloop=10000,
        )
        x_lines_m.append(xx)
        z_lines_m.append(zz)

        _, _, _, xx, _, zz = geopack.trace(
            x_gsm,
            y_gsm,
            z_gsm,
            dir=1,
            rlim=100,
            r0=0.99999,
            parmod=2,
            exname="t89",
            inname="igrf",
            maxloop=10000,
        )
        x_lines_p.append(xx)
        z_lines_p.append(zz)

    for xx, zz in zip(x_lines_m, z_lines_m):
        ax.plot(xx, zz, color="k", linewidth=0.2)

    for xx, zz in zip(x_lines_p, z_lines_p):
        ax.plot(xx, zz, color="k", linewidth=0.2)


def plot_magnetosphere(
    ax,
    tint,
    colors: list = None,
    field_lines: bool = True,
):
    r"""Plot magnetopause, bow shock and earth.

    Parameters
    ----------
    ax : matplotlib.pyplot.subplotsaxes
        Axis to plot.
    tint : list of str
        Time interval.
    colors : list, Optional
        Colors of the magnetopause and the bow show.
        Default use ["tab:blue", "tab:red"]

    Returns
    -------
    ax : matplotlib.pyplot.subplotsaxes
        Axis.

    """

    # Compute Magnetopause
    if colors is None:
        colors = ["tab:blue", "tab:green"]

    x_mp, y_mp = magnetosphere("mp_shue1998", tint)

    # Compute bow show
    x_bs, y_bs = magnetosphere("bs", tint)

    # Plot
    ax.plot(
        np.hstack([x_mp, np.flip(x_mp)]),
        np.hstack([y_mp, np.flip(-y_mp)]),
        color=colors[0],
        label="Magnetopause",
    )
    ax.plot(
        np.hstack([x_bs, np.flip(x_bs)]),
        np.hstack([y_bs, np.flip(-y_bs)]),
        color=colors[1],
        label="Bow Shock",
    )
    _add_earth(ax)

    if field_lines:
        logging.info("Computing field lines using T89 model...")
        _add_field_lines(ax, tint)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from .ts_scalar import ts_scalar

# Local imports
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def remove_repeated_points(inp):
    r"""Remove repeated elements in DataArray or structure data. Important
    when using defatt products. Must have a time variable.

    Parameters
    ----------
    inp : xarray.DataArray or dict
        Time series of the input variable.

    Returns
    -------
    out: xarray.DataArray or dict
        Time series of the cleaned input variable.

    """

    # Points separated in time by less than 100ns are treated as repeats
    threshold = 100

    if isinstance(inp, xr.DataArray):
        diffs = np.diff(inp.time.data.astype(np.int64) * 1e-9)

        no_repeat = np.ones(len(inp))
        no_repeat[diffs < threshold] = 0

        new_time = inp.time.data[no_repeat == 1]
        new_inp = inp.data[no_repeat == 1, :]

        if new_inp.ndim == 1:
            new_data = ts_scalar(new_time, new_inp)
        elif new_inp.ndim == 2:
            new_data = ts_vec_xyz(new_time, new_inp)
        elif new_inp.ndim == 3:
            new_data = ts_vec_xyz(new_time, new_inp)
        else:
            raise TypeError("Invalid data dimension")

    elif isinstance(inp, dict) and ("time" in inp):
        if inp["time"].dtype == "<M8[ns]":
            diffs = np.diff(inp["time"].view("i8") * 1e-9)
        else:
            diffs = np.diff(inp["time"])

        no_repeat = np.ones(len(inp["time"]))
        no_repeat[diffs < threshold] = 0

        var_names = inp.keys()

        for var_name in var_names:
            inp[var_name] = inp[var_name][no_repeat == 1, :]

        new_data = inp
    else:
        # no change to input if it's not a DataArray or structure
        new_data = inp

    return new_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import builtins

# Built-in imports
import datetime
import itertools
import random
import unittest
from unittest import mock

# 3rd party imports
import numpy as np
import xarray as xr
from ddt import data, ddt, idata, unpack

from pyrfu import pyrf
from pyrfu.pyrf.compress_cwt import _compress_cwt_1d
from pyrfu.pyrf.ebsp import _average_data, _censure_plot, _freq_int
from pyrfu.pyrf.int_sph_dist import _mc_cart_2d, _mc_cart_3d, _mc_pol_1d
from pyrfu.pyrf.wavelet import _power_c, _power_r, _ww

from . import generate_data, generate_timeline, generate_ts, generate_vdf

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


class AutoCorrTestCase(unittest.TestCase):
    def test_autocorr_input_type(self):
        self.assertIsNotNone(pyrf.autocorr(generate_ts(64.0, 100, "scalar")))
        self.assertIsNotNone(pyrf.autocorr(generate_ts(64.0, 100, "scalar"), 25))
        self.assertIsNotNone(pyrf.autocorr(generate_ts(64.0, 100, "scalar"), 25, True))

    def test_autocorr_input_shape(self):
        self.assertIsNotNone(pyrf.autocorr(generate_ts(64.0, 100, "scalar")))
        self.assertIsNotNone(pyrf.autocorr(generate_ts(64.0, 100, "vector")))

    def test_autocorr_input_values(self):
        with self.assertRaises(ValueError):
            pyrf.autocorr(generate_ts(64.0, 100, "scalar"), 100)

    def test_autocorr_output_type(self):
        self.assertIsInstance(
            pyrf.autocorr(generate_ts(64.0, 100, "scalar")), xr.DataArray
        )
        self.assertIsInstance(
            pyrf.autocorr(generate_ts(64.0, 100, "vector")), xr.DataArray
        )

    def test_autocorr_output_shape(self):
        result = pyrf.autocorr(generate_ts(64.0, 100, "scalar"))
        self.assertEqual(result.ndim, 1)
        self.assertEqual(result.shape[0], 100)

        result = pyrf.autocorr(generate_ts(64.0, 100, "scalar"), 25)
        self.assertEqual(result.ndim, 1)
        self.assertEqual(result.shape[0], 26)

        result = pyrf.autocorr(generate_ts(64.0, 100, "vector"))
        self.assertEqual(result.ndim, 2)
        self.assertEqual(result.shape[0], 100)
        self.assertEqual(result.shape[1], 3)


class AverageVDFTestCase(unittest.TestCase):
    def test_average_vdf_input_type(self):
        with self.assertRaises(AssertionError):
            pyrf.average_vdf(0, 3)
            pyrf.average_vdf(np.random.random((100, 32, 32, 16)), 3)
            pyrf.average_vdf(generate_vdf(64.0, 100, [32, 32, 16]), [3, 5])

    def test_average_vdf_values(self):
        with self.assertRaises(AssertionError):
            pyrf.average_vdf(generate_vdf(64.0, 100, [32, 32, 16]), 2)

    def test_average_vdf_output_type(self):
        self.assertIsInstance(
            pyrf.average_vdf(generate_vdf(64.0, 100, [32, 32, 16]), 3), xr.Dataset
        )

    def test_average_vdf_output_meta(self):
        avg_inds = np.arange(1, 99, 3, dtype=int)
        result = pyrf.average_vdf(generate_vdf(64.0, 100, [32, 32, 16]), 3)

        self.assertIsInstance(result.attrs["delta_energy_plus"], np.ndarray)
        self.assertEqual(result.attrs["delta_energy_plus"].ndim, 2)
        self.assertEqual(len(result.attrs["delta_energy_plus"]), len(avg_inds))

        self.assertIsInstance(result.attrs["delta_energy_minus"], np.ndarray)
        self.assertEqual(result.attrs["delta_energy_minus"].ndim, 2)
        self.assertEqual(len(result.attrs["delta_energy_minus"]), len(avg_inds))


class Avg4SCTestCase(unittest.TestCase):
    def test_avg_4sc_input(self):
        self.assertIsNotNone(
            pyrf.avg_4sc(
                [
                    generate_ts(64.0, 100, "scalar"),
                    generate_ts(64.0, 100, "scalar"),
                    generate_ts(64.0, 100, "scalar"),
                    generate_ts(64.0, 100, "scalar"),
                ]
            )
        )
        self.assertIsNotNone(
            pyrf.avg_4sc(
                [
                    generate_ts(64.0, 100, "vector"),
                    generate_ts(64.0, 100, "vector"),
                    generate_ts(64.0, 100, "vector"),
                    generate_ts(64.0, 100, "vector"),
                ]
            )
        )
        self.assertIsNotNone(
            pyrf.avg_4sc(
                [
                    generate_ts(64.0, 100, "tensor"),
                    generate_ts(64.0, 100, "tensor"),
                    generate_ts(64.0, 100, "tensor"),
                    generate_ts(64.0, 100, "tensor"),
                ]
            )
        )

        with self.assertRaises(TypeError):
            pyrf.avg_4sc(
                [
                    generate_data(100, "tensor"),
                    generate_data(100, "tensor"),
                    generate_data(100, "tensor"),
                    generate_data(100, "tensor"),
                ]
            )

    def test_avg_4sc_output(self):
        result = pyrf.avg_4sc(
            [
                generate_ts(64.0, 100, "tensor"),
                generate_ts(64.0, 100, "tensor"),
                generate_ts(64.0, 100, "tensor"),
                generate_ts(64.0, 100, "tensor"),
            ]
        )

        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3, 3])


class C4GradTestCase(unittest.TestCase):
    def test_c_4_grad_input(self):
        with self.assertRaises(AssertionError):
            pyrf.c_4_grad(
                generate_ts(64.0, 100, kind="vector"),
                generate_ts(64.0, 100, kind="vector"),
            )
            pyrf.c_4_grad([], [])

            pyrf.c_4_grad(
                [generate_ts(64.0, 100, kind="vector") for _ in range(4)],
                [generate_ts(64.0, 100, kind="vector") for _ in range(4)],
                0,
            )

            pyrf.c_4_grad(
                [generate_ts(64.0, 100, kind="vector") for _ in range(4)],
                [generate_ts(64.0, 100, kind="vector") for _ in range(4)],
                "bazinga",
            )

    def test_c_4_grad_output(self):
        r_mms = [generate_ts(64.0, 100, kind="vector") for _ in range(4)]
        b_mms = [generate_ts(64.0, 100, kind="vector") for _ in range(4)]
        n_mms = [generate_ts(64.0, 100, kind="scalar") for _ in range(4)]

        result = pyrf.c_4_grad(r_mms, b_mms, "grad")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3, 3])

        result = pyrf.c_4_grad(r_mms, b_mms, "div")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(
            list(result.shape),
            [
                100,
            ],
        )

        result = pyrf.c_4_grad(r_mms, b_mms, "curl")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.c_4_grad(r_mms, b_mms, "bdivb")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.c_4_grad(r_mms, b_mms, "curv")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.c_4_grad(r_mms, n_mms, "grad")
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])


class C4JTestCase(unittest.TestCase):
    def test_c_4_j_input(self):
        with self.assertRaises(AssertionError):
            pyrf.c_4_j(
                generate_ts(64.0, 100, kind="vector"),
                generate_ts(64.0, 100, kind="vector"),
            )
            pyrf.c_4_j([], [])

    def test_c_4_j_output(self):
        r_mms = [generate_ts(64.0, 100, kind="vector") for _ in range(4)]
        b_mms = [generate_ts(64.0, 100, kind="vector") for _ in range(4)]
        j, div_b, b_avg, jxb, div_t_shear, div_pb = pyrf.c_4_j(r_mms, b_mms)

        self.assertIsInstance(j, xr.DataArray)
        self.assertListEqual(list(j.shape), [100, 3])

        self.assertIsInstance(div_b, xr.DataArray)
        self.assertListEqual(
            list(div_b.shape),
            [
                100,
            ],
        )

        self.assertIsInstance(b_avg, xr.DataArray)
        self.assertListEqual(list(b_avg.shape), [100, 3])

        self.assertIsInstance(jxb, xr.DataArray)
        self.assertListEqual(list(jxb.shape), [100, 3])

        self.assertIsInstance(div_t_shear, xr.DataArray)
        self.assertListEqual(list(div_t_shear.shape), [100, 3])

        self.assertIsInstance(div_pb, xr.DataArray)
        self.assertListEqual(list(div_pb.shape), [100, 3])


class CalcAgTestCase(unittest.TestCase):
    def test_calc_ag_input_type(self):
        self.assertIsNotNone(pyrf.calc_ag(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_ag(0.0)
            pyrf.calc_ag(generate_data(100))

    def test_calc_ag_output_type(self):
        result = pyrf.calc_ag(generate_ts(64.0, 100, "tensor"))

        # Output must be a xarray
        self.assertIsInstance(result, xr.DataArray)

    def test_calc_ag_output_shape(self):
        result = pyrf.calc_ag(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)

    def test_calc_ag_dims(self):
        result = pyrf.calc_ag(generate_ts(64.0, 100, "tensor"))
        self.assertListEqual(list(result.dims), ["time"])

    def test_calc_ag_meta(self):
        result = pyrf.calc_ag(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.attrs["TENSOR_ORDER"], 0)


class CalcAgyroTestCase(unittest.TestCase):
    def test_calc_agyro_input_type(self):
        self.assertIsNotNone(pyrf.calc_agyro(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_agyro(0.0)
            pyrf.calc_agyro(generate_data(100))

    def test_calc_agyro_output_type(self):
        result = pyrf.calc_agyro(generate_ts(64.0, 100, "tensor"))

        # Output must be a xarray
        self.assertIsInstance(result, xr.DataArray)

    def test_calc_agyro_output_shape(self):
        result = pyrf.calc_agyro(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)

    def test_calc_agyro_dims(self):
        result = pyrf.calc_agyro(generate_ts(64.0, 100, "tensor"))
        self.assertListEqual(list(result.dims), ["time"])

    def test_calc_agyro_meta(self):
        result = pyrf.calc_agyro(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.attrs["TENSOR_ORDER"], 0)


class CalcDngTestCase(unittest.TestCase):
    def test_calc_dng_input_type(self):
        self.assertIsNotNone(pyrf.calc_dng(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_dng(0.0)
            pyrf.calc_dng(generate_data(100))

    def test_calc_dng_output_type(self):
        result = pyrf.calc_dng(generate_ts(64.0, 100, "tensor"))

        # Output must be a xarray
        self.assertIsInstance(result, xr.DataArray)

    def test_calc_dng_output_shape(self):
        result = pyrf.calc_dng(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)

    def test_calc_dng_dims(self):
        result = pyrf.calc_dng(generate_ts(64.0, 100, "tensor"))
        self.assertListEqual(list(result.dims), ["time"])

    def test_calc_dng_meta(self):
        result = pyrf.calc_dng(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.attrs["TENSOR_ORDER"], 0)


class CalcDtTestCase(unittest.TestCase):
    def test_calc_dt_input_type(self):
        self.assertIsNotNone(pyrf.calc_dt(generate_ts(64.0, 100, "scalar")))
        self.assertIsNotNone(pyrf.calc_dt(generate_ts(64.0, 100, "vector")))
        self.assertIsNotNone(pyrf.calc_dt(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_dt(0)
            pyrf.calc_dt(generate_data(100))

    def test_calc_dt_output_type(self):
        self.assertIsInstance(pyrf.calc_dt(generate_ts(64.0, 100)), float)


class CalcFsTestCase(unittest.TestCase):
    def test_calc_fs_input_type(self):
        self.assertIsNotNone(pyrf.calc_fs(generate_ts(64.0, 100)))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_fs(0)
            pyrf.calc_fs(generate_data(100))

    def test_calc_fs_output_type(self):
        self.assertIsInstance(pyrf.calc_fs(generate_ts(64.0, 100)), float)


class CalcSqrtQTestCase(unittest.TestCase):
    def test_calc_sqrtq_input_type(self):
        self.assertIsNotNone(pyrf.calc_sqrtq(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            # Raises error if input is not a xarray
            pyrf.calc_sqrtq(0.0)
            pyrf.calc_sqrtq(generate_data(100))

    def test_calc_sqrtq_output_type(self):
        result = pyrf.calc_sqrtq(generate_ts(64.0, 100, "tensor"))

        # Output must be a xarray
        self.assertIsInstance(result, xr.DataArray)

    def test_calc_sqrtq_output_shape(self):
        result = pyrf.calc_sqrtq(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)

    def test_calc_sqrtq_dims(self):
        result = pyrf.calc_sqrtq(generate_ts(64.0, 100, "tensor"))
        self.assertListEqual(list(result.dims), ["time"])

    def test_calc_sqrtq_meta(self):
        result = pyrf.calc_sqrtq(generate_ts(64.0, 100, "tensor"))
        self.assertEqual(result.attrs["TENSOR_ORDER"], 0)


class Cart2SphTestCase(unittest.TestCase):
    def test_cart2sph_output(self):
        result = pyrf.cart2sph(1.0, 1.0, 1.0)
        self.assertIsInstance(result[0], np.float64)
        self.assertIsInstance(result[1], np.float64)
        self.assertIsInstance(result[2], np.float64)

        result = pyrf.cart2sph(
            np.random.random(100), np.random.random(100), np.random.random(100)
        )
        self.assertIsInstance(result[0], np.ndarray)
        self.assertListEqual(
            list(result[0].shape),
            [
                100,
            ],
        )
        self.assertIsInstance(result[1], np.ndarray)
        self.assertListEqual(
            list(result[1].shape),
            [
                100,
            ],
        )
        self.assertIsInstance(result[2], np.ndarray)
        self.assertListEqual(
            list(result[2].shape),
            [
                100,
            ],
        )


class Cart2SphTsTestCase(unittest.TestCase):
    def test_cart2sph_ts_input(self):
        with self.assertRaises(AssertionError):
            pyrf.cart2sph_ts(0.0)
            pyrf.cart2sph_ts(generate_data(100, "vector"))
            pyrf.cart2sph_ts(generate_ts(64.0, 100, "scalar"))
            pyrf.cart2sph_ts(generate_ts(64.0, 100, "vector"), 2)

    def test_cart2sph_ts_output(self):
        result = pyrf.cart2sph_ts(generate_ts(64.0, 100, "vector"), 1)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.cart2sph_ts(generate_ts(64.0, 100, "vector"), -1)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])


class CdfEpoch2Datetime64TestCase(unittest.TestCase):
    def test_cdfepoch2datetime64_input_type(self):
        ref_time = 599572869184000000
        self.assertIsNotNone(pyrf.cdfepoch2datetime64(ref_time))
        time_line = np.arange(ref_time, int(ref_time + 100))
        self.assertIsNotNone(pyrf.cdfepoch2datetime64(time_line))
        self.assertIsNotNone(pyrf.cdfepoch2datetime64(list(time_line)))

    def test_cdfepoch2datetime64_output_type(self):
        ref_time = 599572869184000000
        self.assertIsInstance(pyrf.cdfepoch2datetime64(ref_time), np.ndarray)
        time_line = np.arange(ref_time, int(ref_time + 100))
        self.assertIsInstance(pyrf.cdfepoch2datetime64(time_line), np.ndarray)
        self.assertIsInstance(pyrf.cdfepoch2datetime64(list(time_line)), np.ndarray)

    def test_cdfepoch2datetime64_output_shape(self):
        ref_time = 599572869184000000
        self.assertEqual(len(pyrf.cdfepoch2datetime64(ref_time)), 1)
        time_line = np.arange(ref_time, int(ref_time + 100))
        self.assertEqual(len(pyrf.cdfepoch2datetime64(time_line)), 100)
        self.assertEqual(len(pyrf.cdfepoch2datetime64(list(time_line))), 100)


@ddt
class CompressCwtTestCase(unittest.TestCase):
    @data(([], 10), (np.random.random((100, 100)), 100))
    @unpack
    def test_compress_cwt_input(self, cwt, nc):
        with self.assertRaises(AssertionError):
            pyrf.compress_cwt(cwt, nc)

    def test_compress_cwt_output(self):
        times = generate_timeline(64.0, 1000)
        freqs = np.logspace(0, 3, 100)
        cwt_x = xr.DataArray(
            np.random.random((1000, 100)), coords=[times, freqs], dims=["time", "f"]
        )
        cwt_y = xr.DataArray(
            np.random.random((1000, 100)), coords=[times, freqs], dims=["time", "f"]
        )
        cwt_z = xr.DataArray(
            np.random.random((1000, 100)), coords=[times, freqs], dims=["time", "f"]
        )
        cwt = xr.Dataset({"x": cwt_x, "y": cwt_y, "z": cwt_z})
        result = pyrf.compress_cwt(cwt, 10)
        self.assertIsInstance(result[0], np.ndarray)

        self.assertIsInstance(result[1], np.ndarray)
        self.assertIsInstance(result[2], np.ndarray)

    def test_compress_cwt_1d(self):
        result = _compress_cwt_1d.__wrapped__(
            np.random.random((1000, 100)), random.randint(2, 100)
        )
        self.assertIsInstance(result, np.ndarray)


class ConvertFACTestCase(unittest.TestCase):
    def test_convert_fac_input(self):
        with self.assertRaises(AssertionError):
            pyrf.convert_fac(0, 0)
            pyrf.convert_fac(generate_data(100, "vector"), generate_data(100, "vector"))

        with self.assertRaises(TypeError):
            pyrf.convert_fac(
                generate_ts(64.0, 100, "tensor"), generate_ts(64.0, 100, "vector")
            )

        with self.assertRaises(TypeError):
            pyrf.convert_fac(
                generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "scalar")
            )

    def test_convert_fac_output(self):
        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "vector"), generate_ts(64.0, 98, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            np.random.random(3),
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 97, "vector"),
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.convert_fac(
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 2])


@ddt
class CotransTestCase(unittest.TestCase):
    @data(
        (0.0, "gse>gsm", True),
        (generate_data(100), "gse>gsm", True),
        (generate_ts(64.0, 100, "vector"), "gsm", True),
        (generate_ts(64.0, 100, "tensor"), "gse>gsm", True),
        (
            generate_ts(64.0, 100, "vector", {"COORDINATE_SYSTEM": "gse"}),
            "gsm>sm",
            True,
        ),
    )
    @unpack
    def test_cotrans_input(self, inp, flag, hapgood):
        with self.assertRaises((TypeError, IndexError, ValueError, AssertionError)):
            pyrf.cotrans(inp, flag, hapgood)

    @idata(itertools.product(["gei", "geo", "gse", "gsm", "mag", "sm"], repeat=2))
    def test_cotrans_output_trans(self, value):
        transf = f"{value[0]}>{value[1]}"

        inp = generate_ts(64.0, 100, "vector")
        result = pyrf.cotrans(inp, transf, hapgood=True)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.cotrans(inp, transf, hapgood=False)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        inp.attrs["COORDINATE_SYSTEM"] = value[0]
        result = pyrf.cotrans(inp, transf, hapgood=False)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.cotrans(inp, value[1], hapgood=False)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.cotrans(inp, value[1], hapgood=True)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

    def test_cotrans_output_exot(self):
        inp = generate_ts(64.0, 100, "scalar")
        result = pyrf.cotrans(inp, "gse>gsm", hapgood=True)
        self.assertIsInstance(result, xr.DataArray)
        result = pyrf.cotrans(inp, "dipoledirectiongse", hapgood=True)
        self.assertIsInstance(result, xr.DataArray)


class CrossTestCase(unittest.TestCase):
    def test_cross_input(self):
        with self.assertRaises(AssertionError):
            pyrf.cross(
                generate_ts(64.0, 100, "scalar"), generate_ts(64.0, 100, "scalar")
            )
            pyrf.cross(
                generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "scalar")
            )
            pyrf.cross(
                generate_ts(64.0, 100, "scalar"), generate_ts(64.0, 100, "vector")
            )

    def test_cross_output(self):
        result = pyrf.cross(
            generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])

        result = pyrf.cross(
            generate_ts(64.0, 100, "vector"), generate_ts(64.0, 97, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])


@ddt
class DateStrTestCase(unittest.TestCase):
    def test_date_str_input(self):
        with self.assertRaises(AssertionError):
            pyrf.date_str("2019-01-01T00:00:00")
            pyrf.date_str([np.datetime64("2019-01-01T00:00:00"), "2019-01-01T00:10:00"])
            pyrf.date_str(["2019-01-01T00:00:00", "2019-01-01T00:10:00"], 1)

            tint = ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]
            pyrf.date_str(tint, 0)
            pyrf.date_str(tint, 5)

    @idata(range(1, 5))
    def test_date_str_output(self, value):
        tint = ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]
        result = pyrf.date_str(tint, value)
        self.assertIsInstance(result, str)


class Datetime2Iso8601TestCase(unittest.TestCase):
    def test_datetime2iso8601_input_type(self):
        ref_time = datetime.datetime(2019, 1, 1, 0, 0, 0, 0)
        time_line = [ref_time + datetime.timedelta(seconds=i) for i in range(10)]
        self.assertIsNotNone(pyrf.datetime2iso8601(ref_time))
        self.assertIsNotNone(pyrf.datetime2iso8601(time_line))

    def test_datetime2iso8601_output_type(self):
        ref_time = datetime.datetime(2019, 1, 1, 0, 0, 0, 0)
        time_line = [ref_time + datetime.timedelta(seconds=i) for i in range(10)]
        self.assertIsInstance(pyrf.datetime2iso8601(ref_time), str)
        self.assertIsInstance(pyrf.datetime2iso8601(time_line), list)

    def test_datetime2iso8601_output_shape(self):
        ref_time = datetime.datetime(2019, 1, 1, 0, 0, 0, 0)
        time_line = [ref_time + datetime.timedelta(seconds=i) for i in range(10)]

        # ISO8601 contains 29 characters (nanosecond precision)
        self.assertEqual(len(pyrf.datetime2iso8601(ref_time)), 29)
        self.assertEqual(len(pyrf.datetime2iso8601(time_line)), 10)


@ddt
class Datetime642Iso8601TestCase(unittest.TestCase):
    @data(
        datetime.datetime(2019, 1, 1, 0, 0, 0),
        "2019-01-01T00:00:00.000000000",
    )
    def test_datetime642iso8601_input(self, value):
        with self.assertRaises(TypeError):
            pyrf.datetime642iso8601(value)

    @data(np.datetime64("2019-01-01T00:00:00.000000000"), generate_timeline(64.0, 100))
    def test_datetime642iso8601_output(self, value):
        self.assertIsInstance(pyrf.datetime642iso8601(value), np.ndarray)


@ddt
class Datetime642TtnsTestCase(unittest.TestCase):
    @data(
        datetime.datetime(2019, 1, 1, 0, 0, 0),
        "2019-01-01T00:00:00.000000000",
    )
    def test_datetime642ttns_input(self, value):
        with self.assertRaises(TypeError):
            pyrf.datetime642ttns(value)

    @data(np.datetime64("2019-01-01T00:00:00.000000000"), generate_timeline(64.0, 100))
    def test_datetime642ttns_output(self, value):
        self.assertIsInstance(pyrf.datetime642ttns(value), np.ndarray)


@ddt
class Datetime642UnixTestCase(unittest.TestCase):
    @data(
        datetime.datetime(2019, 1, 1, 0, 0, 0),
        "2019-01-01T00:00:00.000000000",
    )
    def test_datetime642unix_input(self, value):
        with self.assertRaises(TypeError):
            pyrf.datetime642unix(value)

    @data(np.datetime64("2019-01-01T00:00:00.000000000"), generate_timeline(64.0, 100))
    def test_datetime642unix_output(self, value):
        self.assertIsInstance(pyrf.datetime642unix(value), np.ndarray)


@ddt
class DecParPerpTestCase(unittest.TestCase):
    @data(
        (generate_data(100, "vector"), generate_ts(64.0, 100, "vector"), False),
        (generate_ts(64.0, 100, "vector"), generate_data(100, "vector"), False),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), 0),
        (generate_ts(64.0, 100, "scalar"), generate_ts(64.0, 100, "vector"), False),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "scalar"), False),
    )
    @unpack
    def test_dec_par_perp_input(self, inp, b_bgd, flag_spin_plane):
        with self.assertRaises(AssertionError):
            pyrf.dec_par_perp(inp, b_bgd, flag_spin_plane)

    @data(
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), False),
        (
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector") * 1e-4,
            False,
        ),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), True),
    )
    @unpack
    def test_dec_par_perp_output(self, inp, b_bgd, flag_spin_plane):
        a_para, a_perp, alpha = pyrf.dec_par_perp(inp, b_bgd, flag_spin_plane)
        self.assertIsInstance(a_para, xr.DataArray)
        self.assertIsInstance(a_perp, xr.DataArray)


@ddt
class DistAppendTestCase(unittest.TestCase):
    @data(
        (None, generate_vdf(64.0, 100, [32, 32, 16])),
        (generate_vdf(64.0, 100, [32, 32, 16]), generate_vdf(64.0, 100, [32, 32, 16])),
    )
    @unpack
    def test_dist_append_output(self, inp0, inp1):
        result = pyrf.dist_append(inp0, inp1)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class DynamicPressTestCase(unittest.TestCase):
    @data(
        (generate_data(100, "scalar"), generate_ts(64.0, 100, "vector"), "ions"),
        (generate_ts(64.0, 100, "scalar"), generate_data(100, "vector"), "ions"),
    )
    @unpack
    def test_dynamic_press_input(self, n_s, v_xyz, specie):
        with self.assertRaises(AssertionError):
            pyrf.dynamic_press(n_s, v_xyz, specie)

    @data("ions", "electrons")
    def test_dynamic_press_output(self, value):
        result = pyrf.dynamic_press(
            generate_ts(64.0, 100, "scalar"), generate_ts(64.0, 100, "vector"), value
        )
        self.assertIsInstance(result, xr.DataArray)
        self.assertEqual(result.ndim, 1)


@ddt
class EVxBTestCase(unittest.TestCase):
    @data(
        (generate_data(100, "vector"), generate_ts(64.0, 100, "vector"), "vxb"),
        (generate_ts(64.0, 100, "vector"), generate_data(100, "vector"), "vxb"),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), "bazinga"),
    )
    @unpack
    def test_e_vxb_input(self, v_xyz, b_xyz, flag):
        with self.assertRaises((TypeError, AssertionError)):
            pyrf.e_vxb(v_xyz, b_xyz, flag)

    @data(
        (generate_ts(64.0, 100, "vector"), "vxb"),
        (generate_ts(64.0, 100, "vector"), "exb"),
        (np.random.random(3), "vxb"),
        (np.random.random(3), "exb"),
    )
    @unpack
    def test_e_vxb_output(self, v_xyz, flag):
        result = pyrf.e_vxb(v_xyz, generate_ts(64.0, 100, "vector"), flag)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), [100, 3])


@ddt
class EbNRFTestCase(unittest.TestCase):
    @data("a", "b", np.random.random(3))
    def test_eb_nrf_output(self, value):
        result = pyrf.eb_nrf(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            value,
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EdbTestCase(unittest.TestCase):
    @data("e.b=0", "e_perp+nan", "e_par")
    def test_edb_output(self, value):
        pyrf.edb(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            random.random() * 90,
            value,
        )


@ddt
class EbspTestCase(unittest.TestCase):
    @data(
        (
            None,
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
            {},
        ),
        (
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 98, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 120, "vector"),
            [1e0, 1e1],
            {},
        ),
        (
            generate_ts(99.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
            {},
        ),
        (
            generate_ts(40.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(40.0, 100, "vector"),
            generate_ts(40.0, 100, "vector"),
            generate_ts(40.0, 100, "vector"),
            [1e0, 1e1],
            {},
        ),
        (
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            [1e0, 1e1],
            {},
        ),
        (
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            generate_ts(64.0, 97, "vector"),
            [1e0, 1e1],
            {"fac_matrix": generate_ts(64.0, 100, "tensor")},
        ),
        (
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            None,
            [1e0, 1e1],
            {},
        ),
    )
    @unpack
    def test_ebsp_input_pass(self, e_xyz, db_xyz, b_xyz, b_bgd, xyz, freq_int, options):
        result = pyrf.ebsp(e_xyz, db_xyz, b_xyz, b_bgd, xyz, freq_int, **options)
        self.assertIsInstance(result, dict)

    @data(
        (
            generate_ts(64.0, 100, "vector"),
            None,
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
        ),
        (
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            None,
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
        ),
        (
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            None,
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
        ),
        (
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
        ),
        (
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e1, 1e0],
        ),
        (
            generate_ts(64.0, 100, "vector")[:, :2],
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
        ),
    )
    @unpack
    def test_ebsp_input_fail(self, e_xyz, db_xyz, b_xyz, b_bgd, xyz, freq_int):
        with self.assertRaises((AssertionError, TypeError, IndexError, ValueError)):
            pyrf.ebsp(e_xyz, db_xyz, b_xyz, b_bgd, xyz, freq_int)

    @data(
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": True,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": True,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": False,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": False,
            "de_dot_b0": True,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": True,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": True,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": random.randint(2, 50),
            "fac_matrix": None,
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": True,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": generate_ts(64.0, 100, "tensor"),
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": generate_ts(64.0, 100, "tensor"),
            "m_width_coeff": 1,
        },
        {
            "polarization": False,
            "no_resample": False,
            "fac": True,
            "de_dot_b0": False,
            "full_b_db": False,
            "nav": 8,
            "fac_matrix": None,
            "m_width_coeff": random.random(),
        },
    )
    def test_ebsp_options(self, value):
        result = pyrf.ebsp(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            [1e0, 1e1],
            **value,
        )

        self.assertIsInstance(result, dict)
        self.assertIsInstance(result["bb_xxyyzzss"], xr.DataArray)

    @data("pc12", "pc35", [1e0, 1e1])
    def test_ebsp_freq_int_pass(self, value):
        self.assertIsNotNone(
            pyrf.ebsp(
                generate_ts(64.0, 100000, "vector"),
                generate_ts(64.0, 100000, "vector"),
                generate_ts(64.0, 100000, "vector"),
                generate_ts(64.0, 100000, "vector"),
                generate_ts(64.0, 100000, "vector"),
                value,
            )
        )

    @data(random.random(), np.random.random(3), "bazinga", [1, 100])
    def test_ebsp_freq_int_fail(self, value):
        with self.assertRaises((AssertionError, ValueError)):
            pyrf.ebsp(
                generate_ts(64.0, 10000, "vector"),
                generate_ts(64.0, 10000, "vector"),
                generate_ts(64.0, 10000, "vector"),
                generate_ts(64.0, 10000, "vector"),
                generate_ts(64.0, 10000, "vector"),
                value,
            )

    @data(([0.32, 3.2], generate_ts(64.0, 100000, "vector")))
    @unpack
    def test_average_data(self, freq_int, data):
        _, _, _, out_time = _freq_int(freq_int, data)
        in_time = data.time.data.astype(np.float64) / 1e9

        result = _average_data.__wrapped__(data.data, in_time, out_time, None)
        self.assertIsInstance(result, np.ndarray)
        self.assertListEqual(list(result.shape), [len(out_time), 3])

    @data(([0.32, 3.2], generate_ts(64.0, 100000, "vector")))
    @unpack
    def test_censure_plot(self, freq_int, data):
        _, _, out_sampling, out_time = _freq_int(freq_int, data)
        a_ = np.logspace(1, 2, 12)
        idx_nan = np.full(len(data), False)
        idx_nan[np.random.randint(100000, size=(100))] = True
        censure = np.floor(2 * a_ * out_sampling / 64.0 * 8)
        result = _censure_plot.__wrapped__(
            np.random.random((len(out_time), len(a_))),
            idx_nan,
            censure,
            len(data),
            a_,
        )
        self.assertIsInstance(result, np.ndarray)
        self.assertListEqual(list(result.shape), [len(out_time), len(a_)])


class EndTestCase(unittest.TestCase):
    def test_end_input(self):
        with self.assertRaises(AssertionError):
            pyrf.end(generate_timeline(64.0, 100))

    def test_end_output(self):
        pyrf.end(generate_ts(64.0, 100))


@ddt
class EstimateTestCase(unittest.TestCase):
    @data(
        ("bazinga", random.random(), None),
        ("capacitance_wire", 0, random.random()),
        ("capacitance_wire", random.randint(1, 9), random.randint(1, 9)),
        ("capacitance_cylinder", random.randint(20, 100), random.randint(1, 9)),
    )
    @unpack
    def test_estimate_input(self, what_to_estimate, radius, length):
        with self.assertRaises((NotImplementedError, ValueError)):
            pyrf.estimate(what_to_estimate, radius, length)

    @data(
        ("capacitance_disk", random.random(), None),
        ("capacitance_sphere", random.random(), None),
        ("capacitance_wire", random.random(), random.randint(10, 100)),
        ("capacitance_cylinder", random.randint(1, 9), random.randint(40, 100)),
        ("capacitance_cylinder", random.randint(1, 9), random.randint(5, 26)),
    )
    @unpack
    def test_estimate_output(self, what_to_estimate, radius, length):
        result = pyrf.estimate(what_to_estimate, radius, length)
        self.assertIsInstance(result, float)


@ddt
class ExtendTintTestCase(unittest.TestCase):
    def test_extend_tint_input(self):
        with self.assertRaises(TypeError):
            pyrf.extend_tint([0, 0], None)

    @data(
        (
            ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"],
            [-random.random(), random.random()],
        ),
        (["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"], None),
        (
            [
                np.datetime64("2019-01-01T00:00:00.000000000"),
                np.datetime64("2019-01-01T00:10:00.000000000"),
            ],
            None,
        ),
    )
    @unpack
    def test_extend_tint_ouput(self, tint, ext):
        pyrf.extend_tint(tint, ext)


@ddt
class FiltTestCase(unittest.TestCase):
    @data(
        (generate_data(100), 0, random.randint(1, 22), random.choice(range(1, 10, 2))),
        (
            generate_ts(64.0, 100),
            "bazinga",
            random.randint(1, 22),
            random.choice(range(1, 10, 2)),
        ),
        (
            generate_ts(64.0, 100),
            random.randint(1, 22),
            "bazinga",
            random.choice(range(1, 10, 2)),
        ),
        (generate_ts(64.0, 100), 0, random.randint(1, 22), "ORDEEERRRR"),
    )
    @unpack
    def test_filt_input(self, inp, f_min, f_max, order):
        with self.assertRaises(AssertionError):
            pyrf.filt(inp, f_min, f_max, order)

    @data(
        (
            generate_ts(64.0, 100, "scalar"),
            0,
            1,
            -1,
        ),
        (
            generate_ts(64.0, 100, "vector"),
            0,
            1,
            -1,
        ),
        (
            generate_ts(64.0, 100, "vector"),
            0,
            random.randint(2, 22),
            -1,
        ),
        (
            generate_ts(64.0, 100, "vector"),
            0,
            random.randint(2, 22),
            random.choice(range(1, 10, 2)),
        ),
        (
            generate_ts(64.0, 100, "vector"),
            random.randint(2, 22),
            0,
            -1,
        ),
        (
            generate_ts(64.0, 100, "vector"),
            random.randint(2, 22),
            0,
            random.choice(range(1, 10, 2)),
        ),
        (
            generate_ts(64.0, 100, "vector"),
            random.randint(2, 11),
            random.randint(12, 22),
            -1,
        ),
        (
            generate_ts(64.0, 100, "vector"),
            random.randint(2, 11),
            random.randint(12, 22),
            random.choice(range(1, 10, 2)),
        ),
    )
    @unpack
    def test_filt_output(self, inp, f_min, f_max, order):
        result = pyrf.filt(inp, f_min, f_max, order)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class GradientTestCase(unittest.TestCase):
    @data(
        generate_ts(64.0, 100, "scalar"),
        generate_ts(64.0, 100, "vector"),
        generate_ts(64.0, 100, "tensor"),
    )
    def test_gradient_output(self, value):
        value.attrs = {"UNITS": "bazinga"}
        result = pyrf.gradient(value)
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(list(result.shape), list(value.shape))


@ddt
class Gse2GsmTestCase(unittest.TestCase):
    @data(
        (generate_data(100, "vector"), "gse>gsm"),
        (generate_ts(64.0, 100, "scalar"), "gse>gsm"),
        (generate_ts(64.0, 100, "vector"), "bazinga"),
    )
    @unpack
    def test_gse2gsm_input(self, inp, flag):
        with self.assertRaises(AssertionError):
            pyrf.gse2gsm(inp, flag)

    @data(
        (generate_ts(64.0, 100, "vector"), "gse>gsm"),
        (generate_ts(64.0, 100, "vector"), "gsm>gse"),
    )
    @unpack
    def test_gse2gsm_output(self, inp, flag):
        pyrf.gse2gsm(inp, flag)


@ddt
class HistogramTestCase(unittest.TestCase):
    @data(
        (random.randint(2, 100), None, None, None),
        (np.sort(np.random.random(10)), None, None, None),
        ("fd", None, None, None),
        ("auto", np.sort(np.random.random(2)), None, None),
        (100, None, np.random.random(1000), None),
        ("auto", None, None, True),
    )
    @unpack
    def test_histogram_output(self, bins, y_range, weights, density):
        result = pyrf.histogram(
            generate_ts(64.0, 1000), bins, y_range, weights, density
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class Histogram2DTestCase(unittest.TestCase):
    @data(
        (random.randint(2, 100), None, None, None),
        (np.sort(np.random.random(100)), None, None, None),
        (np.random.randint(2, 100, size=(2,)), None, None, None),
        ([np.sort(np.random.random(100)) for _ in range(2)], None, None, None),
        (100, np.sort(np.random.random((2, 2)), axis=1), None, None),
        (100, None, np.random.random(1000), None),
        (100, None, None, True),
    )
    @unpack
    def test_histogram2d_output(self, bins, y_range, weights, density):
        result = pyrf.histogram2d(
            generate_ts(64.0, 1000),
            generate_ts(64.0, 900),
            bins,
            y_range,
            weights,
            density,
        )
        self.assertIsInstance(result, xr.DataArray)
        result = pyrf.histogram2d(
            generate_ts(64.0, 1000),
            generate_ts(64.0, 1000),
            bins,
            y_range,
            weights,
            density,
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class IncrementsTestCase(unittest.TestCase):
    @data(
        generate_ts(64.0, 100, "scalar"),
        generate_ts(64.0, 100, "vector"),
        generate_ts(64.0, 100, "tensor"),
    )
    def test_increments_output(self, value):
        result = pyrf.increments(value, random.randint(1, 50))
        self.assertIsInstance(result[0], np.ndarray)
        self.assertIsInstance(result[1], xr.DataArray)


@ddt
class IntSphDistTestCase(unittest.TestCase):
    @data(
        {"projection_base": "pol", "projection_dim": "2d"},
    )
    def test_int_sph_dist_input(self, value):
        vdf = np.random.random((51, 32, 16))
        speed = np.linspace(0, 1, 51)
        phi = np.arange(32)
        theta = np.arange(16)
        speed_grid = np.linspace(-1, 1, 101)

        with self.assertRaises((RuntimeError, NotImplementedError)):
            pyrf.int_sph_dist(vdf, speed, phi, theta, speed_grid, **value)

    @data(
        {},
        {"weight": "lin"},
        {"weight": "log"},
        {"speed_edges": np.linspace(-0.01, 1.01, 52)},
        {"speed_grid_edges": np.linspace(-1.01, 1.01, 102)},
        {
            "phi_grid": np.arange(0, 32),
            "projection_base": "cart",
            "projection_dim": "2d",
        },
        {"projection_base": "cart", "projection_dim": "3d"},
    )
    def test_int_sph_dist_output(self, value):
        vdf = np.random.random((51, 32, 16))
        speed = np.linspace(0, 1, 51)
        phi = np.arange(32)
        theta = np.arange(16)
        speed_grid = np.linspace(-1, 1, 101)
        result = pyrf.int_sph_dist(vdf, speed, phi, theta, speed_grid, **value)
        self.assertIsInstance(result, dict)

    @data(
        (
            np.random.random((51, 32, 16)),
            np.linspace(0, 1, 51),
            np.arange(32),
            np.arange(16),
            np.ones(51) * 0.02,
            np.ones(51) * 0.01,
            np.ones(32),
            np.ones(16),
            np.linspace(-1.01, 1.01, 102),
            np.ones(101) * 0.02 * np.pi / 16,
            np.array([-np.inf, np.inf]),
            np.array([-np.pi, np.pi]),
            np.ones((51, 32, 16), dtype=np.int64) * 10,
            np.eye(3),
        )
    )
    def test_mc_pol_1d(self, value):
        vdf, *args = value
        vdf[vdf < 1e-2] = 0
        self.assertIsInstance(_mc_pol_1d.__wrapped__(vdf, *args), np.ndarray)

    @data(
        (
            np.random.random((51, 32, 16)),
            np.linspace(0, 1, 51),
            np.arange(32),
            np.arange(16),
            np.ones(51) * 0.02,
            np.ones(51) * 0.01,
            np.ones(32),
            np.ones(16),
            np.linspace(-1.01, 1.01, 102),
            0.02**2,
            np.array([-np.inf, np.inf]),
            np.array([-np.pi, np.pi]),
            (np.ones((51, 32, 16), dtype=np.int64) * 10).astype(int),
            np.eye(3),
        )
    )
    def test_mc_cart_2d(self, value):
        vdf, *args = value
        vdf[vdf < 1e-2] = 0
        self.assertIsInstance(_mc_cart_2d.__wrapped__(vdf, *args), np.ndarray)

    @data(
        (
            np.random.random((51, 32, 16)),
            np.linspace(0, 1, 51),
            np.arange(32),
            np.arange(16),
            np.ones(51) * 0.02,
            np.ones(51) * 0.01,
            np.ones(32),
            np.ones(16),
            np.linspace(-1.01, 1.01, 102),
            0.02**2,
            np.array([-np.inf, np.inf]),
            np.array([-np.pi, np.pi]),
            (np.ones((51, 32, 16), dtype=np.int64) * 10).astype(int),
            np.eye(3),
        )
    )
    def test_mc_cart_3d(self, value):
        vdf, *args = value
        vdf[vdf < 1e-2] = 0
        self.assertIsInstance(_mc_cart_3d.__wrapped__(vdf, *args), np.ndarray)


@ddt
class IntegrateTestCase(unittest.TestCase):
    @data(generate_ts(64.0, 100, "scalar"), generate_ts(64.0, 100, "vector"))
    def test_integrate_output(self, value):
        result = pyrf.integrate(value)
        self.assertIsInstance(result, xr.DataArray)


class IPlasmaCalcTestCase(unittest.TestCase):
    def test_iplasma_calc_output(self):
        with mock.patch.object(builtins, "input", lambda _: random.randint(10, 100)):
            result = pyrf.iplasma_calc(True, True)
            self.assertIsInstance(result, dict)

            result = pyrf.iplasma_calc(False, False)
            self.assertIsNone(result)


@ddt
class Iso86012DatetimeTestCase(unittest.TestCase):
    @data(
        "2019-01-01T00:00:00.000000000",
        ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"],
        np.array(["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]),
    )
    def test_iso86012datetime(self, value):
        result = pyrf.iso86012datetime(value)
        self.assertIsInstance(result, list)


@ddt
class Iso86012Unix(unittest.TestCase):
    @data(
        "2019-01-01T00:00:00.000000000",
        ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"],
        np.array(["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]),
    )
    def test_iso86012unix_output(self, value):
        result = pyrf.iso86012unix(value)
        self.assertIsInstance(result, np.ndarray)


@ddt
class Iso86012TimeVec(unittest.TestCase):
    @data(
        "2019-01-01T00:00:00.000000000",
        ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"],
        np.array(["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]),
    )
    def test_iso86012timevec_output(self, value):
        result = pyrf.iso86012timevec(value)
        self.assertIsInstance(result, np.ndarray)


@ddt
class LowPassTestCase(unittest.TestCase):
    @data(
        generate_ts(64.0, 10000, "scalar"),
        generate_ts(64.0, 10000, "vector"),
        generate_ts(64.0, 10000, "tensor"),
    )
    def test_lowpass_output(self, value):
        pyrf.lowpass(value, random.random(), 64.0)


@ddt
class LShellTestCase(unittest.TestCase):
    @data("gei", "geo", "gse", "gsm", "mag", "sm")
    def test_l_shell_output(self, value):
        result = pyrf.l_shell(
            generate_ts(64.0, 100, "vector", {"COORDINATE_SYSTEM": value})
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class MeanTestCase(unittest.TestCase):
    @data(None, generate_ts(64.0, 100, "vector"))
    def test_mean_output(self, value):
        result = pyrf.mean(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "vector"),
            value,
        )
        self.assertIsInstance(result, xr.DataArray)


class MeanBinsTestCase(unittest.TestCase):
    def test_mean_bins_output(self):
        result = pyrf.mean_bins(
            generate_ts(64.0, 100), generate_ts(64.0, 100), random.randint(2, 20)
        )
        self.assertIsInstance(result, xr.Dataset)


class MedianBinsTestCase(unittest.TestCase):
    def test_median_bins_output(self):
        result = pyrf.median_bins(
            generate_ts(64.0, 100), generate_ts(64.0, 100), random.randint(2, 20)
        )
        self.assertIsInstance(result, xr.Dataset)


@ddt
class MvaTestCase(unittest.TestCase):
    @data("mvar", "<bn>=0", "td")
    def test_mva_output(self, method):
        result = pyrf.mva(generate_ts(64.0, 100, "vector"), method)
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], np.ndarray)
        self.assertIsInstance(result[2], np.ndarray)


@ddt
class NewXyzTestCase(unittest.TestCase):
    @data(generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "tensor"))
    def test_new_xyz_output(self, inp):
        result = pyrf.new_xyz(inp, np.random.random((3, 3)))
        self.assertIsInstance(result, xr.DataArray)
        self.assertEqual(result.ndim, inp.ndim)


class NormTestCase(unittest.TestCase):
    def test_norm_output(self):
        result = pyrf.norm(generate_ts(64.0, 100, "vector"))
        self.assertIsInstance(result, xr.DataArray)
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)


@ddt
class PlasmaBetaTestCase(unittest.TestCase):
    @data((generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "tensor")))
    @unpack
    def test_plasma_beta_output(self, b_xyz, p_xyz):
        result = pyrf.plasma_beta(b_xyz, p_xyz)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class StructFuncTestCase(unittest.TestCase):
    @data(
        (generate_ts(64.0, 100, "scalar"), None, random.randint(1, 100)),
        (generate_ts(64.0, 100, "vector"), None, random.randint(1, 100)),
        (generate_ts(64.0, 100, "tensor"), None, random.randint(1, 100)),
        (generate_ts(64.0, 100, "vector"), np.random.randint([1] * 50, [50] * 50), 1),
    )
    @unpack
    def test_struct_func_output(self, inp, scales, order):
        result = pyrf.struct_func(inp, scales, order)
        self.assertIsInstance(result, xr.DataArray)


class TraceTestCase(unittest.TestCase):
    def test_trace_input(self):
        self.assertIsNotNone(pyrf.trace(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            pyrf.trace(generate_data(100, "tensor"))
            pyrf.trace(generate_ts(64.0, 100, "scalar"))
            pyrf.trace(generate_ts(64.0, 100, "vector"))

    def test_trace_output(self):
        result = pyrf.trace(generate_ts(64.0, 100, "tensor"))
        self.assertIsInstance(result, xr.DataArray)
        self.assertListEqual(
            list(result.shape),
            [
                100,
            ],
        )


class OptimizeNbins1DTestCase(unittest.TestCase):
    def test_optimize_nbins_1d(self):
        result = pyrf.optimize_nbins_1d(
            generate_ts(64.0, 1000),
            n_min=random.randint(2, 10),
            n_max=random.randint(20, 100),
        )

        self.assertIsInstance(result, int)


class OptimizeNbins2DTestCase(unittest.TestCase):
    def test_optimize_nbins_2d(self):
        result = pyrf.optimize_nbins_2d(
            generate_ts(64.0, 1000),
            generate_ts(64.0, 1000),
            n_min=[random.randint(2, 10), random.randint(2, 10)],
            n_max=[random.randint(20, 100), random.randint(20, 100)],
        )
        self.assertIsInstance(result[0], int)
        self.assertIsInstance(result[1], int)


class Pid4SCTestCase(unittest.TestCase):
    def test_pid_4sc_output(self):
        result = pyrf.pid_4sc(
            [generate_ts(64.0, 100, "vector") for _ in range(4)],
            [generate_ts(64.0, 100, "vector") for _ in range(4)],
            [generate_ts(64.0, 100, "tensor") for _ in range(4)],
            [generate_ts(64.0, 100, "vector") for _ in range(4)],
        )
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], xr.DataArray)


class PlasmaCalcTestCase(unittest.TestCase):
    def test_plasma_calc_output(self):
        result = pyrf.plasma_calc(
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
        )
        self.assertIsInstance(result, xr.Dataset)


@ddt
class ResampleTestCase(unittest.TestCase):
    @data(
        (generate_ts(64.0, 100), generate_ts(640.0, 1000)),
        (generate_ts(640.0, 1000), generate_ts(64.0, 100)),
        (generate_vdf(64.0, 100, [32, 32, 16]), generate_ts(640.0, 1000)),
        (generate_ts(64.0, 100), generate_ts(640.0, 2)),
        (generate_ts(64.0, 100), generate_ts(640.0, 1)),
    )
    @unpack
    def test_resample_output(self, inp, ref):
        result = pyrf.resample(inp, ref)
        self.assertIsInstance(result, type(inp))


@ddt
class PoyntingFluxTestCase(unittest.TestCase):
    @data(
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector")),
        (generate_ts(128.0, 200, "vector"), generate_ts(64.0, 100, "vector")),
        (generate_ts(64.0, 100, "vector"), generate_ts(128.0, 200, "vector")),
    )
    @unpack
    def test_poynting_flux_output(self, e_xyz, b_xyz):
        result = pyrf.poynting_flux(e_xyz, b_xyz, None)
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], xr.DataArray)

        result = pyrf.poynting_flux(e_xyz, b_xyz, generate_ts(64.0, 100, "vector"))
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], xr.DataArray)
        self.assertIsInstance(result[2], xr.DataArray)


class PresAnisTestCase(unittest.TestCase):
    def test_pres_anis_output(self):
        result = pyrf.pres_anis(
            generate_ts(64.0, 100, "tensor"), generate_ts(64.0, 100, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class ShockNormalTestCase(unittest.TestCase):
    def test_shock_normal_input(self):
        with self.assertRaises(AssertionError):
            pyrf.shock_normal([])

        with self.assertRaises(TypeError):
            pyrf.shock_normal(
                {
                    "b_u": np.random.random((3)),
                    "b_d": np.random.random((3)),
                    "v_u": np.random.random((3)),
                    "v_d": np.random.random((3)),
                    "n_u": random.random(),
                    "n_d": random.random(),
                    "r_xyz": random.random(),
                }
            )

    @data(
        {
            "b_u": np.random.random((3)),
            "b_d": np.random.random((3)),
            "v_u": np.random.random((3)),
            "v_d": np.random.random((3)),
            "n_u": random.random(),
            "n_d": random.random(),
        },
        {
            "b_u": np.random.random((2, 3)),
            "b_d": np.random.random((2, 3)),
            "v_u": np.random.random((2, 3)),
            "v_d": np.random.random((2, 3)),
            "n_u": np.random.random((2, 1)),
            "n_d": np.random.random((2, 1)),
        },
        {
            "b_u": np.random.random((3)),
            "b_d": np.random.random((3)),
            "v_u": np.random.random((3)),
            "v_d": np.random.random((3)),
            "n_u": random.random(),
            "n_d": random.random(),
            "r_xyz": np.random.random((3)),
        },
        {
            "b_u": np.random.random((3)),
            "b_d": np.random.random((3)),
            "v_u": np.random.random((3)),
            "v_d": np.random.random((3)),
            "n_u": random.random(),
            "n_d": random.random(),
            "r_xyz": generate_ts(64.0, 100, "vector"),
        },
        {
            "b_u": np.random.random((3)),
            "b_d": np.random.random((3)),
            "v_u": np.random.random((3)),
            "v_d": np.random.random((3)),
            "n_u": random.random(),
            "n_d": random.random(),
            "r_xyz": generate_ts(64.0, 100, "vector"),
            "d2u": random.choice([-1, 1]),
            "dt_f": random.random(),
            "f_cp": random.random(),
        },
    )
    def test_shock_normal_ouput(self, value):
        result = pyrf.shock_normal(value)
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result["v_sh"], dict)


@ddt
class ShockParametersTestCase(unittest.TestCase):
    def test_shock_parameters_input(self):
        with self.assertRaises(AssertionError):
            pyrf.shock_parameters(
                {
                    "b": np.random.random(3),
                    "n": random.random(),
                    "v": np.random.random(3),
                    "t_i": random.random(),
                    "t_e": random.random(),
                    "v_sh": random.random(),
                    "nvec": np.random.random(3),
                    "ref_sys": "bazinga",
                }
            )

    @data(
        {
            "b": np.random.random(3),
            "n": random.random(),
            "v": np.random.random(3),
            "t_i": random.random(),
            "t_e": random.random(),
            "ref_sys": "nif",
        },
        {
            "b": np.random.random(3),
            "n": random.random(),
            "v": np.random.random(3),
            "t_i": random.random(),
            "t_e": random.random(),
            "v_sh": random.random(),
            "nvec": np.random.random(3),
            "ref_sys": "nif",
        },
        {
            "b": np.random.random(3),
            "n": random.random(),
            "v": np.random.random(3),
            "t_i": random.random(),
            "t_e": random.random(),
            "v_sh": random.random(),
            "nvec": np.random.random(3),
            "ref_sys": "sc",
        },
    )
    def test_shock_parameters_output(self, value):
        pyrf.shock_parameters(value)


@ddt
class SolidAngleTestCase(unittest.TestCase):
    @data(
        tuple(np.random.random(3) for _ in range(3)),
        tuple(generate_data(100, "vector") for _ in range(3)),
        tuple(generate_ts(64.0, 100, "vector") for _ in range(3)),
    )
    @unpack
    def test_solid_angle_ouput(self, inp0, inp1, inp2):
        result = pyrf.solid_angle(inp0, inp1, inp2)
        self.assertIsInstance(result, np.ndarray)


@ddt
class Sph2CartTestCase(unittest.TestCase):
    @data(
        tuple(generate_data(100, "scalar") for _ in range(3)),
        tuple(generate_ts(64.0, 100, "scalar") for _ in range(3)),
    )
    @unpack
    def test_sph2cart_output(self, azimuth, elevation, r):
        self.assertIsNotNone(pyrf.sph2cart(azimuth, elevation, r))


class StartTestCase(unittest.TestCase):
    def test_start_input_type(self):
        self.assertIsNotNone(pyrf.start(generate_ts(64.0, 100, "scalar")))
        self.assertIsNotNone(pyrf.start(generate_ts(64.0, 100, "vector")))
        self.assertIsNotNone(pyrf.start(generate_ts(64.0, 100, "tensor")))

        with self.assertRaises(AssertionError):
            pyrf.start(0)
            pyrf.start(generate_timeline(64.0, 100))

    def test_start_output(self):
        result = pyrf.start(generate_ts(64.0, 100, "scalar"))
        self.assertIsInstance(result, np.float64)
        self.assertEqual(
            np.datetime64(int(result * 1e9), "ns"),
            np.datetime64("2019-01-01T00:00:00.000"),
        )


@ddt
class TsAppendTestCase(unittest.TestCase):
    @data(
        generate_ts(64.0, 100, "scalar"),
        generate_ts(64.0, 100, "vector"),
        generate_ts(64.0, 100, "tensor"),
    )
    def test_ts_append_output(self, value):
        value.attrs = {
            "bazinga": "This is my spot!!",
            "I AM GROOT": "I AM STEVE ROGERS",
            "random": np.random.random(100),
        }
        value.time.attrs = {
            "bazinga": "This is my spot!!",
            "I AM GROOT": "I AM STEVE ROGERS",
            "random": np.random.random(100),
        }

        result = pyrf.ts_append(None, value)
        self.assertIsInstance(result, xr.DataArray)
        self.assertEqual(result.ndim, value.ndim)

        result = pyrf.ts_append(value, value)
        self.assertIsInstance(result, xr.DataArray)
        self.assertEqual(result.ndim, value.ndim)


@ddt
class TimeClipTestCase(unittest.TestCase):
    @data(
        [
            datetime.datetime(2019, 1, 1, 0, 0, 0, 312),
            datetime.datetime(2019, 1, 1, 0, 0, 0, 468),
        ],
        "2019-01-01T00:00:00.312",
    )
    def test_time_clip_input(self, value):
        with self.assertRaises(TypeError):
            pyrf.time_clip(generate_ts(64.0, 100), value)

    @data(generate_ts(64.0, 100), generate_vdf(64.0, 100, (32, 32, 16)))
    def test_time_clip_output(self, value):
        result = pyrf.time_clip(
            value, ["2019-01-01T00:00:00.312", "2019-01-01T00:00:00.468"]
        )
        self.assertIsInstance(result, type(value))

        result = pyrf.time_clip(
            value,
            [
                np.datetime64("2019-01-01T00:00:00.312"),
                np.datetime64("2019-01-01T00:00:00.468"),
            ],
        )
        self.assertIsInstance(result, type(value))

        result = pyrf.time_clip(value, generate_ts(64.0, 20))
        self.assertIsInstance(result, type(value))


class TsSkymapTestCase(unittest.TestCase):
    def test_ts_skymap_input_type(self):
        with self.assertRaises(AssertionError):
            pyrf.ts_skymap(0, 0, 0, 0, 0)

    def test_ts_skymap_output_type(self):
        result = pyrf.ts_skymap(
            generate_timeline(64.0, 100),
            np.random.random((100, 32, 32, 16)),
            np.random.random((100, 32)),
            np.random.random((100, 32)),
            np.random.random(16),
        )
        self.assertIsInstance(result, xr.Dataset)

    def test_ts_skymap_output_shape(self):
        result = pyrf.ts_skymap(
            generate_timeline(64.0, 100),
            np.random.random((100, 32, 32, 16)),
            np.random.random((100, 32)),
            np.random.random((100, 32)),
            np.random.random(16),
        )
        self.assertEqual(result.data.ndim, 4)
        self.assertListEqual(list(result.data.shape), [100, 32, 32, 16])
        self.assertEqual(result.energy.ndim, 2)
        self.assertListEqual(list(result.energy.shape), [100, 32])
        self.assertEqual(result.phi.ndim, 2)
        self.assertListEqual(list(result.phi.shape), [100, 32])
        self.assertEqual(result.theta.ndim, 1)
        self.assertListEqual(list(result.theta.shape), [16])

    def test_ts_skymap_output_meta(self):
        result = pyrf.ts_skymap(
            generate_timeline(64.0, 100),
            np.random.random((100, 32, 32, 16)),
            np.random.random((100, 32)),
            np.random.random((100, 32)),
            np.random.random(16),
        )
        self.assertListEqual(
            list(result.attrs.keys()), ["energy0", "energy1", "esteptable"]
        )
        self.assertListEqual(
            list(result.attrs["energy0"].shape),
            [
                32,
            ],
        )
        self.assertListEqual(
            list(result.attrs["energy1"].shape),
            [
                32,
            ],
        )
        self.assertListEqual(
            list(result.attrs["esteptable"].shape),
            [
                100,
            ],
        )

        for k in result:
            self.assertEqual(result[k].attrs, {})


class TsScalarTestCase(unittest.TestCase):
    def test_ts_scalar_input_type(self):
        with self.assertRaises(AssertionError):
            pyrf.ts_scalar(0, 0)
            pyrf.ts_scalar(
                list(generate_timeline(64.0, 100)), list(generate_data(100, "scalar"))
            )

    def test_ts_scalar_input_shape(self):
        with self.assertRaises(AssertionError):
            # Raises error if data and timeline don't have the same size
            pyrf.ts_scalar(generate_timeline(64.0, 100), generate_data(99, "scalar"))
            # Raises error if vector as input
            pyrf.ts_scalar(generate_timeline(64.0, 100), generate_data(100, "vector"))
            # Raises error if tensor as input
            pyrf.ts_scalar(generate_timeline(64.0, 100), generate_data(100, "tensor"))

    def test_ts_scalar_output_type(self):
        result = pyrf.ts_scalar(
            generate_timeline(64.0, 100), generate_data(100, "scalar")
        )
        self.assertIsInstance(result, xr.DataArray)

    def test_ts_scalar_output_shape(self):
        result = pyrf.ts_scalar(
            generate_timeline(64.0, 100), generate_data(100, "scalar")
        )
        self.assertEqual(result.ndim, 1)
        self.assertEqual(len(result), 100)

    def test_ts_scalar_dims(self):
        result = pyrf.ts_scalar(
            generate_timeline(64.0, 100), generate_data(100, "scalar")
        )
        self.assertListEqual(list(result.dims), ["time"])

    def test_ts_scalar_meta(self):
        result = pyrf.ts_scalar(
            generate_timeline(64.0, 100), generate_data(100, "scalar")
        )
        self.assertEqual(result.attrs["TENSOR_ORDER"], 0)


@ddt
class TsSpectrTestCase(unittest.TestCase):
    @data(
        (0, np.random.random(10), np.random.random((100, 10)), "energy", None),
        (generate_timeline(64.0, 100), 0, np.random.random((100, 10)), "energy", None),
        (generate_timeline(64.0, 100), 0, np.random.random((100, 10)), "energy", None),
        (
            generate_timeline(64.0, 100),
            np.random.random(10),
            np.random.random((98, 10)),
            "energy",
            None,
        ),
        (
            generate_timeline(64.0, 100),
            np.random.random(10),
            np.random.random((100, 9)),
            "energy",
            None,
        ),
    )
    @unpack
    def test_ts_spectr_input(self, time, ener, data, comp_name, attrs):
        with self.assertRaises(AssertionError):
            pyrf.ts_spectr(time, ener, data, comp_name, attrs)

    def test_ts_spectr_output(self):
        result = pyrf.ts_spectr(
            generate_timeline(64.0, 100),
            np.random.random(10),
            np.random.random((100, 10)),
        )
        self.assertIsInstance(result, xr.DataArray)


class TsVecXYZTestCase(unittest.TestCase):
    def test_ts_vec_xyz_input_type(self):
        with self.assertRaises(AssertionError):
            pyrf.ts_vec_xyz(0, 0)
            pyrf.ts_vec_xyz(
                list(generate_timeline(64.0, 100)), list(generate_data(100, "vector"))
            )

    def test_ts_vec_xyz_input_shape(self):
        with self.assertRaises(AssertionError):
            # Raises error if data and timeline don't have the same size
            pyrf.ts_vec_xyz(generate_timeline(64.0, 100), generate_data(99, "vector"))
            # Raises error if vector as input
            pyrf.ts_vec_xyz(generate_timeline(64.0, 100), generate_data(100, "scalar"))
            # Raises error if tensor as input
            pyrf.ts_vec_xyz(generate_timeline(64.0, 100), generate_data(100, "tensor"))

    def test_ts_vec_xyz_output_type(self):
        result = pyrf.ts_vec_xyz(
            generate_timeline(64.0, 100), generate_data(100, "vector")
        )
        self.assertIsInstance(result, xr.DataArray)

    def test_ts_vec_xyz_output_shape(self):
        result = pyrf.ts_vec_xyz(
            generate_timeline(64.0, 100), generate_data(100, "vector")
        )
        self.assertEqual(result.ndim, 2)
        self.assertEqual(result.shape[0], 100)
        self.assertEqual(result.shape[1], 3)

    def test_ts_vec_xyz_dims(self):
        result = pyrf.ts_vec_xyz(
            generate_timeline(64.0, 100), generate_data(100, "vector")
        )
        self.assertListEqual(list(result.dims), ["time", "comp"])

    def test_ts_vec_xyz_meta(self):
        result = pyrf.ts_vec_xyz(
            generate_timeline(64.0, 100), generate_data(100, "vector")
        )
        self.assertEqual(result.attrs["TENSOR_ORDER"], 1)


class TsTensorXYZTestCase(unittest.TestCase):
    def test_ts_tensor_xyz_input_type(self):
        with self.assertRaises(AssertionError):
            pyrf.ts_tensor_xyz(0, 0)
            pyrf.ts_tensor_xyz(
                list(generate_timeline(64.0, 100)), list(generate_data(100, "tensor"))
            )

    def test_ts_tensor_xyz_input_shape(self):
        with self.assertRaises(AssertionError):
            # Raises error if data and timeline don't have the same size
            pyrf.ts_tensor_xyz(
                generate_timeline(64.0, 100), generate_data(99, "tensor")
            )
            # Raises error if vector as input
            pyrf.ts_tensor_xyz(
                generate_timeline(64.0, 100), generate_data(100, "scalar")
            )
            # Raises error if tensor as input
            pyrf.ts_tensor_xyz(
                generate_timeline(64.0, 100), generate_data(100, "vector")
            )

    def test_ts_tensor_xyz_output_type(self):
        result = pyrf.ts_tensor_xyz(
            generate_timeline(64.0, 100), generate_data(100, "tensor")
        )
        self.assertIsInstance(result, xr.DataArray)

    def test_ts_tensor_xyz_output_shape(self):
        result = pyrf.ts_tensor_xyz(
            generate_timeline(64.0, 100), generate_data(100, "tensor")
        )
        self.assertEqual(result.ndim, 3)
        self.assertEqual(result.shape[0], 100)
        self.assertEqual(result.shape[1], 3)
        self.assertEqual(result.shape[2], 3)

    def test_ts_tensor_xyz_dims(self):
        result = pyrf.ts_tensor_xyz(
            generate_timeline(64.0, 100), generate_data(100, "tensor")
        )
        self.assertListEqual(list(result.dims), ["time", "comp_h", "comp_v"])

    def test_ts_tensor_xyz_meta(self):
        result = pyrf.ts_tensor_xyz(
            generate_timeline(64.0, 100), generate_data(100, "tensor")
        )
        self.assertEqual(result.attrs["TENSOR_ORDER"], 2)


@ddt
class Ttns2Datetime64TestCase(unittest.TestCase):
    @data(
        int(random.random() * 1e12),
        [int(random.random() * 1e12), int(random.random() * 1e12)],
        np.array([int(random.random() * 1e12), int(random.random() * 1e12)]),
    )
    def test_ttns2datetime64_output(self, value):
        result = pyrf.ttns2datetime64(value)
        self.assertIsInstance(result, np.ndarray)


@ddt
class WaveletTestCase(unittest.TestCase):
    @data(
        (generate_ts(64.0, 100, "tensor"), {}),
        (generate_ts(64, 100, "vector"), {"linear": [random.randint(10, 100)]}),
    )
    @unpack
    def test_wavelet_input(self, inp, options):
        with self.assertRaises(TypeError):
            pyrf.wavelet(inp, **options)

    @data(
        (generate_ts(64, 100, "scalar"), {}),
        (generate_ts(64, 99, "scalar"), {}),
        (generate_ts(64, 100, "vector"), {}),
        (generate_ts(64, 100, "vector"), {"linear": True}),
        (generate_ts(64, 100, "vector"), {"linear": random.randint(10, 100)}),
        (
            generate_ts(64, 100, "vector"),
            {"linear": random.randint(10, 100), "return_power": False},
        ),
    )
    @unpack
    def test_wavelet_output(self, inp, options):
        self.assertIsNotNone(pyrf.wavelet(inp, **options))

    @data(
        (
            np.random.random((100, 1)),
            np.random.random((1, 200)),
            random.random(),
            np.random.random((100, 1)),
            random.randint(16, 96),
        )
    )
    @unpack
    def test_ww(self, s_ww, scales_mat, sigma, frequencies_mat, f_nyq):
        self.assertIsNotNone(
            _ww.__wrapped__(s_ww, scales_mat, sigma, frequencies_mat, f_nyq)
        )

    @data(
        (
            np.random.random((100, 3)) + np.random.random((100, 3)) * 1j,
            np.random.random((100, 3)),
        )
    )
    @unpack
    def test_power_r(self, power, new_freq_mat):
        self.assertIsNotNone(_power_r.__wrapped__(power, new_freq_mat))

    @data(
        (
            np.random.random((100, 3)) + np.random.random((100, 3)) * 1j,
            np.random.random((100, 3)),
        )
    )
    @unpack
    def test_power_c(self, power, new_freq_mat):
        self.assertIsNotNone(_power_c.__wrapped__(power, new_freq_mat))


@ddt
class VhtTestCase(unittest.TestCase):
    @data(
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), True),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 103, "vector"), True),
        (generate_ts(64.0, 100, "vector"), generate_ts(64.0, 100, "vector"), False),
    )
    @unpack
    def test_vht_output(self, e, b, no_ez):
        result = pyrf.vht(e, b, no_ez)

        self.assertIsInstance(result[0], np.ndarray)
        self.assertIsInstance(result[1], xr.DataArray)
        self.assertIsInstance(result[2], np.ndarray)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_split_integral_ch(inp_dataset):
    r"""This function splits the last integral channel from the FEEPS spectra,
    creating 2 new DataArrays

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Energetic particles energy spectrum from FEEPS.

    Returns
    -------
    out : xarray.Dataset
        Energetic particles energy spectra with the integral channel removed.
    out_500kev : xarray.Dataset
        Integral channel that was removed.

    """

    out_dict, out_dict_500kev = [{}, {}]

    out_dict["spinsectnum"] = inp_dataset["spinsectnum"]

    for k in inp_dataset:
        try:
            # Energy spectra with the integral channel removed
            out_dict[k] = inp_dataset[k][:, :-1]

            # Integral channel that was removed
            out_dict_500kev[k] = inp_dataset[k][:, -1]
        except IndexError:
            pass

    out = xr.Dataset(out_dict, attrs=inp_dataset.attrs)

    out_500kev = xr.Dataset(out_dict_500kev, attrs=inp_dataset.attrs)

    return out, out_500kev



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from pyrfu import pyrf

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


__all__ = [
    "generate_timeline",
    "generate_data",
    "generate_ts",
    "generate_spectr",
    "generate_vdf",
]


def generate_timeline(f_s, n_pts: int = 10000, dtype="datetime64[ns]"):
    r"""Generate timeline for testings

    Parameters
    ----------
    f_s
    n_pts
    dtype

    Returns
    -------

    """
    ref_time = np.datetime64("2019-01-01T00:00:00.000000000")
    times = [ref_time + np.timedelta64(int(i * 1e9 / f_s), "ns") for i in range(n_pts)]
    times = np.array(times).astype(dtype)
    return times


def generate_data(n_pts, kind: str = "scalar"):
    r"""Generate data (numpy.ndarray) for testings

    Parameters
    ----------
    n_pts
    kind

    Returns
    -------

    """
    if kind == "scalar":
        data = np.random.random((n_pts,))
    elif kind == "vector":
        data = np.random.random((n_pts, 3))
    elif kind == "tensor":
        data = np.random.random((n_pts, 3, 3))
    else:
        raise ValueError("Invalid kind of data!!")

    return data


def generate_ts(f_s, n_pts, kind: str = "scalar", attrs=None):
    r"""Generate timeseries for testings

    Parameters
    ----------
    f_s
    n_pts
    kind
    attrs

    Returns
    -------

    """
    if attrs is None:
        attrs = {}

    if kind == "scalar":
        out = pyrf.ts_scalar(
            generate_timeline(f_s, n_pts), generate_data(n_pts, kind), attrs=attrs
        )
    elif kind == "vector":
        out = pyrf.ts_vec_xyz(
            generate_timeline(f_s, n_pts), generate_data(n_pts, kind), attrs=attrs
        )
    elif kind == "tensor":
        out = pyrf.ts_tensor_xyz(
            generate_timeline(f_s, n_pts), generate_data(n_pts, kind), attrs=attrs
        )
    else:
        raise ValueError("Invalid kind of data!!")

    return out


def generate_spectr(f_s, n_pts, shape, attrs=None):
    r"""Generates spectrum for testings

    Parameters
    ----------
    f_s
    n_pts
    shape
    attrs

    Returns
    -------

    """
    out = pyrf.ts_spectr(
        generate_timeline(f_s, n_pts),
        np.arange(shape),
        np.random.random((n_pts, shape)),
        attrs=attrs,
    )
    return out


def generate_vdf(
    f_s,
    n_pts,
    shape,
    energy01: bool = False,
    species: str = "ions",
    units: str = "s^3/cm^6",
):
    r"""Generate skymap for testings

    Parameters
    ----------
    f_s
    n_pts
    shape
    energy01
    species

    Returns
    -------

    """
    times = generate_timeline(f_s, n_pts)

    phi = np.arange(shape[1])
    phi = np.tile(phi, (n_pts, 1))
    theta = np.arange(shape[2])
    data = np.random.random((n_pts, *shape))

    if energy01:
        energy0 = np.linspace(0, shape[0], shape[0], endpoint=False)
        energy1 = np.linspace(0, shape[0], shape[0], endpoint=False) + 0.5
        esteptable = np.arange(n_pts) % 2
        energy = np.tile(energy0, (n_pts, 1))
        energy[esteptable == 1, :] = np.tile(energy1, (np.sum(esteptable), 1))
    else:
        energy = np.linspace(0, shape[0], shape[0], endpoint=False)
        energy = np.tile(energy, (n_pts, 1))
        energy0 = energy[0, :]
        energy1 = energy[1, :]
        esteptable = np.zeros(n_pts)

    attrs = {"UNITS": units}
    glob_attrs = {
        "species": species,
        "delta_energy_plus": np.ones((n_pts, shape[0])),
        "delta_energy_minus": np.ones((n_pts, shape[0])),
    }

    out = pyrf.ts_skymap(
        times,
        data,
        energy,
        phi,
        theta,
        energy0=energy0,
        energy1=energy1,
        esteptable=esteptable,
        attrs=attrs,
        glob_attrs=glob_attrs,
    )
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .cross import cross
from .dot import dot

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def c_4_k(r_list):
    r"""Calculates reciprocal vectors in barycentric coordinates.

    Parameters
    ----------
    r_list : list of xarray.DataArray
        Position of the spacecrafts.

    Returns
    -------
    k_list : list of xarray.DataArray
        Reciprocal vectors in barycentric coordinates.

    Notes
    -----
    The units of reciprocal vectors are the same as [1/r].

    """

    mms_list = np.arange(4)

    k_list = [r_list[0].copy()] * 4

    mms_list_r0 = np.roll(mms_list, 0)
    mms_list_r1 = np.roll(mms_list, 1)
    mms_list_r2 = np.roll(mms_list, 2)
    mms_list_r3 = np.roll(mms_list, 3)

    for i, alpha, beta, gamma in zip(
        mms_list_r0,
        mms_list_r1,
        mms_list_r2,
        mms_list_r3,
    ):
        dr_jk_x_dr_jm = cross(
            r_list[beta] - r_list[alpha],
            r_list[gamma] - r_list[alpha],
        )

        dr12 = r_list[i] - r_list[alpha]

        k_list[i] = dr_jk_x_dr_jm / dot(dr_jk_x_dr_jm, dr12)

    return k_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .get_omni_data import get_omni_data

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _rzero(d_p, b_z):
    return (10.22 + 1.29 * np.tanh(0.184 * (b_z + 8.14))) * d_p ** (-1 / 6.6)


def _alpha(d_p, b_z):
    return (0.58 - 0.007 * b_z) * (1 + 0.024 * np.log(d_p))


def magnetosphere(model: str = "mp_shue1998", tint: list = None):
    r"""Returns the location of magnetopause.

    Parameters
    ----------
    model : str
        Model to use. Implemented: 'mp_shue1998', 'bs'.
        Default is 'mp_shue1998'.
    tint : list
        Time interval.

    Returns
    -------
    x_ : ndarray
        X location of the magnetopause.
    y_ : ndarray
        Y location of the magnetopause.


    Examples
    --------
    >>> from pyrfu.pyrf import magnetosphere

    >>> x_mp, y_mp = magnetosphere("mp_shue1998", 10, -2)

    """

    if tint is None:
        d_p = 1.7389104
        b_z = 0.012695087
        m_a = 4
    else:
        omni_data = get_omni_data(["p", "bzgsm", "ma"], tint)
        b_z = float(omni_data.bzgsm.mean("time").data)
        d_p = float(omni_data.p.mean("time").data)
        m_a = float(omni_data.ma.mean("time").data)

    if model == "mp_shue1998":
        theta_ = np.linspace(0, np.pi, int(np.pi / 0.1))
        r_zero = _rzero(d_p, b_z)
        alpha_ = _alpha(d_p, b_z)

        with np.errstate(divide="ignore"):
            r_s = r_zero * (2.0 / (1 + np.cos(theta_))) ** alpha_

        x_s = r_s * np.cos(theta_)
        y_s = r_s * np.sin(theta_)
        y_s = y_s[abs(x_s) < 100]
        x_s = x_s[abs(x_s) < 100]

    elif model == "bs":
        x_mp, _ = magnetosphere("mp_shue1998", tint)
        gamma_ = 5 / 3

        rstandoff = x_mp[0] * (
            1 + 1.1 * ((gamma_ - 1) * m_a**2 + 2) / ((gamma_ + 1) * (m_a**2 - 1))
        )
        # Smaller increments at the subsolar point
        x_s = (
            rstandoff - np.logspace(np.log10(0.1), np.log10(100 + rstandoff), 300) + 0.1
        )
        # original F/G model adds rstandoff^2=645
        y_s = np.sqrt(
            0.04 * (x_s - rstandoff) ** 2 - 45.3 * (x_s - rstandoff),
        )

    else:
        raise NotImplementedError("This model is not implemented yet!!")

    return x_s, y_s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os
import re
import warnings
from bisect import bisect_left
from datetime import datetime, timedelta
from shutil import copy, copyfileobj
from tempfile import NamedTemporaryFile

# 3rd party imports
import numpy as np
import pkg_resources
import requests
import tqdm
from dateutil.parser import parse

# Local imports
from .tokenize import tokenize

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)

LASP_PUBL = "https://lasp.colorado.edu/mms/sdc/public/files/api/v1/"
LASP_SITL = "https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/"


def _login_lasp(user: str, password: str, lasp_url: str):
    r"""Login to LASP colorado."""

    session = requests.Session()
    session.auth = (user, password)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        _ = session.post("https://lasp.colorado.edu", verify=True, timeout=5)
        testget = session.get(lasp_url, verify=True, timeout=5)

    assert testget != "401", "Login failed!!"

    return session, user


def _construct_url(tint, mms_id, var, lasp_url):
    r"""Construct the url that return a json-formatted string of science
    filenames that are available for download according to:
    https://lasp.colorado.edu/mms/sdc/team/about/how-to/
    """

    tint = np.array(tint).astype("<M8[ns]").astype(str)
    tint = [datetime.strptime(t_[:-3], "%Y-%m-%dT%H:%M:%S.%f") for t_ in tint]
    start_date = tint[0].strftime("%Y-%m-%d")
    end_date = (tint[1] - timedelta(seconds=1)).strftime("%Y-%m-%d-%H-%M-%S")

    url = f"{lasp_url}/file_info/science"
    url = f"{url}?start_date={start_date}&end_date={end_date}&sc_id=mms{mms_id}"

    url = f"{url}&instrument_id={var['inst']}"
    url = f"{url}&data_rate_mode={var['tmmode']}"
    url = f"{url}&data_level={var['lev']}"

    if var["dtype"]:
        url = f"{url}&descriptor={var['dtype']}"

    return url


def _files_in_interval(in_files, trange):
    r"""Filters the file list returned by the SDC to the requested time
    range. This filter is purposefully liberal, it regularly grabs an extra
    file due to special cases
    """

    file_name = r"mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    file_times = []

    regex = re.compile(file_name)

    for file in in_files:
        matches = regex.match(file["file_name"])
        if matches:
            file_times.append(
                (
                    file["file_name"],
                    parse(matches.groups()[0]).timestamp(),
                    file["timetag"],
                    file["file_size"],
                ),
            )

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect_left(times, parse(trange[0]).timestamp())

    def mkout(f):
        return {"file_name": f[0], "timetag": f[2], "size": f[3]}

    if idx_min == 0:
        files = list(map(mkout, sorted_files[idx_min:]))
    else:
        files = list(map(mkout, sorted_files[idx_min - 1 :]))

    return files


def _make_path(file, var, mms_id, lasp_url, data_path: str = ""):
    r"""Construct path of the data file using the standard convention."""

    file_date = parse(file["timetag"])

    if not data_path:
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MMS configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    assert os.path.exists(data_path), "local data directory doesn't exist!"

    path_list = [
        data_path,
        f"mms{mms_id}",
        var["inst"],
        var["tmmode"],
        var["lev"],
        var["dtype"],
        *file_date.strftime("%Y-%m").split("-"),
    ]

    if var["tmmode"].lower() == "brst":
        path_list.append(file_date.strftime("%d"))

    out_path = os.path.join(*path_list)
    out_file = os.path.join(*path_list, file["file_name"])

    download_url = f"{lasp_url}download/science?file={file['file_name']}"

    return out_path, out_file, download_url


def download_data(
    var_str, tint, mms_id, login: str = "", password: str = "", data_path: str = ""
):
    r"""Downloads files containing field `var_str` over the time interval
    `tint` for the spacecraft `mms_id`. The files are saved to `data_path`.

    Parameters
    ----------
    var_str : str
        Input key of variable.
    tint : list of str
        Time interval.
    mms_id : str or int
        Index of the target spacecraft.
    login : str, Optional
        Login to LASP MMS SITL. Default downloads from
        https://lasp.colorado.edu/mms/sdc/public/
    password : str, Optional
        Password to LASP MMS SITL. Default downloads from
        https://lasp.colorado.edu/mms/sdc/public/
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu/mms/config.json`

    """

    if not login:
        lasp_url = LASP_PUBL
    else:
        lasp_url = LASP_SITL

    sdc_session, _ = _login_lasp(login, password, lasp_url)

    headers = {}
    try:
        release_version = pkg_resources.get_distribution("pyrfu").version
    except pkg_resources.DistributionNotFound:
        release_version = "bleeding edge"

    headers["User-Agent"] = f"pyrfu {release_version}"

    var = tokenize(var_str)

    root_path = os.path.dirname(os.path.abspath(__file__))

    with open(
        os.sep.join([root_path, "mms_keys.json"]), "r", encoding="utf-8"
    ) as json_file:
        keys_ = json.load(json_file)

    var["dtype"] = keys_[var["inst"]][var_str.lower()]["dtype"]

    url = _construct_url(tint, mms_id, var, lasp_url)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        http_json = sdc_session.get(url, verify=True, headers=headers).json()

    files_in_interval = _files_in_interval(http_json["files"], tint)

    for file in files_in_interval:
        out_path, out_file, dwl_url = _make_path(file, var, mms_id, lasp_url, data_path)

        logging.info("Downloading %s from %s...", os.path.basename(out_file), dwl_url)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ResourceWarning)
            with sdc_session.get(
                dwl_url,
                stream=True,
                verify=True,
                headers=headers,
            ) as fsrc:
                with NamedTemporaryFile(delete=False) as ftmp:
                    with tqdm.tqdm.wrapattr(
                        fsrc.raw,
                        "read",
                        total=file["size"],
                        ncols=60,
                    ) as fsrc_raw:
                        with open(ftmp.name, "wb") as fs:
                            copyfileobj(fsrc_raw, fs)

                os.makedirs(out_path, exist_ok=True)

                # if the download was successful, copy to data directory
                copy(ftmp.name, out_file)

    sdc_session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.resample import resample
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def correct_edp_probe_timing(sc_pot):
    r"""Corrects for the channel delays not accounted for in the MMS EDP
    processing. As described in the MMS EDP data products guide.

    Parameters
    ----------
    sc_pot : xarray.DataArray
        Time series created from L2 sc_pot files, from the variable
        "mms#_edp_dcv_brst_l2" containing individual probe potentials.

    Returns
    -------
    v_corrected : xarray.DataArray
        Time series where the channel delay for each probe have been
        accounted and corrected for.

    Notes
    -----
    This function is only useful for Burst mode data. For the other
    telemetry modes (i.e. slow and fast) the channel delays are
    completely negligible and the interpolation and resampling applied
    here will have no effect other than possibly introduce numerical
    noise.

    """

    e_fact = [0.1200, 0.1200, 0.0292]

    # Reconstruct E12, E34, E56 as computed in MMS processing
    time = sc_pot.time.data

    diff_sc_pot = []
    for i, fact in zip([0, 2, 4], e_fact):
        diff_sc_pot.append(
            ts_scalar(time, np.diff(sc_pot.data[:, i : i + 2]) / fact),
        )

    # Correct the time tags to create individual time series
    tau_vs = [
        np.timedelta64(0, "ns"),
        np.timedelta64(7629, "ns"),
        np.timedelta64(15259, "ns"),
    ]
    tau_es = [
        np.timedelta64(26703, "ns"),
        np.timedelta64(30518, "ns"),
        np.timedelta64(34332, "ns"),
    ]

    # Odds probes potential 1, 3, 5
    sc_pot_odds = []
    for tau, i in zip(tau_vs, [0, 2, 4]):
        sc_pot_odds.append(ts_scalar(time + tau, sc_pot.data[:, i]))

    # Electric field
    diff_sc_pot = []
    for e_, tau in zip(diff_sc_pot, tau_es):
        diff_sc_pot.append(ts_scalar(e_.time.data + tau, e_.data))

    # Resample all data to time tags of V1 (i.e. timeOrig).
    sc_pot_odds = [resample(v, sc_pot) for v in sc_pot_odds]
    diff_sc_pot = [resample(e, sc_pot) for e in diff_sc_pot]

    # Recompute individual even probe potentials 2, 4, 6
    sc_pot_even = []
    for v_, e_, fact in zip(sc_pot_odds, diff_sc_pot, e_fact):
        sc_pot_even.append(v_ - e_ * fact)

    sc_pot_corrected = []
    for tup in zip(sc_pot_even, sc_pot_odds):
        for item in tup:
            sc_pot_corrected.append(item.data)

    # Create the new time series with the corrected values
    options = {"coords": [time, np.arange(1, 7)], "dims": ["time", "probe"]}
    sc_pot_corrected = xr.DataArray(sc_pot_corrected, **options)

    return sc_pot_corrected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

# 3rd party import
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import LogNorm

# Local imports
from .plot_line import plot_line

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_clines(axis, inp, yscale="log", cscale="log", cmap="jet", **kwargs):
    r"""Plot lines with color associated to the level.

    Parameters
    ----------
    axis :
        Axes
    inp : xarray.DataArray
        Time series as an energy spectrum to plot.
    yscale : str, Optional
        Scale of the yaxis. Default is "log"
    cscale : str, Optional
        Scale of the colormap. Default is "log".
    cmap : str, Optional
        Colormap. Default is "jet"
    kwargs : dict
        Plot options.

    Returns
    -------
    axis :
        Updated axis
    cbl :
        Colorbar associated

    Other Parameters
    ----------------
    See pyrfu.plot.plot_line

    """

    pad = 0.01
    c_map = get_cmap(name=cmap)

    for i, c in enumerate(c_map(np.linspace(0, 1, len(inp.energy.data)))):
        plot_line(axis, inp[:, i], color=c, **kwargs)

    pos = axis.get_position()
    f = plt.gcf()
    cax = f.add_axes([pos.x0 + pos.width + pad, pos.y0, pad, pos.height])

    if cscale == "log":
        norm = LogNorm(vmin=inp.energy.data[0], vmax=inp.energy.data[-1])
    else:
        raise NotImplementedError

    ColorbarBase(cax, cmap=c_map, norm=norm, orientation="vertical")

    if yscale == "log":
        axis.set_yscale("log")

    cax.set_axisbelow(False)

    return axis, cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def unix2datetime64(time):
    r"""Converts unix time to datetime64 in ns units.

    Parameters
    ----------
    time : ndarray
        Time in unix format.

    Returns
    -------
    time_datetime64 : ndarray
        Time in datetime64 format.

    See Also
    --------
    pyrfu.pyrf.datetime642unix

    """

    # Make sure that time is in ns format
    time_unix = (time * 1e9).astype(np.int64)

    # Convert to unix
    time_datetime64 = time_unix.astype("datetime64[ns]")

    return time_datetime64



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_projection(
    axis,
    v_x,
    v_y,
    f_mat,
    vlim: float = 1e3,
    clim: list = None,
    cbar_pos: str = "top",
    cmap: str = "jet",
):
    r"""Plot the projection of the distribution.

    Parameters
    ----------
    axis : matplotlib.pyplot.subplotsaxes
        Axis to plot.
    v_x : ndarray
        X axis velocity grid.
    v_y : ndarray
        Y axis velocity grid.
    f_mat : ndarray
        Projected distribution.
    vlim : float, Optional
        Maximum velocity to limit axis. Default is vlim = 1000 km/s.
    clim : list, Optional
        Caxes limit. Default is clim = [-18, -13] (assume to be in SI units)
    cbar_pos : str, Optional
        Location of the colorbar. Default is cbar_pos = "top".
    cmap : str, Optional
        Colormap. Default is cmap = "jet".

    Returns
    -------
    axis : axis
        Axis
    caxis : axis
        Colorbar axis.

    """

    if clim is None:
        clim = [None, None]

    image = axis.pcolormesh(
        v_x / 1e3,
        v_y / 1e3,
        np.log10(f_mat.T),
        cmap=cmap,
        vmin=clim[0],
        vmax=clim[1],
    )
    axis.set_xlim([-vlim / 1e3, vlim / 1e3])
    axis.set_ylim([-vlim / 1e3, vlim / 1e3])
    axis.set_aspect("equal")

    f = plt.gcf()
    pos = axis.get_position()
    if cbar_pos == "top":
        caxis = f.add_axes(
            [pos.x0, pos.y0 + pos.height + 0.01, pos.width, 0.01],
        )
        f.colorbar(
            mappable=image,
            cax=caxis,
            ax=axis,
            orientation="horizontal",
        )
        caxis.xaxis.set_ticks_position("top")
        caxis.xaxis.set_label_position("top")
    elif cbar_pos == "right":
        caxis = f.add_axes(
            [pos.x0 + pos.width + 0.01, pos.y0, 0.01, pos.height],
        )
        f.colorbar(mappable=image, cax=caxis, ax=axis)
    else:
        raise ValueError("invalic position")

    return axis, caxis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def iso86012unix(time):
    r"""Converts time in iso format to unix

    Parameters
    ----------
    time : str or array_like of str
        Time.

    Returns
    -------
    out : float or list of float
        Time in unix format.

    """

    assert isinstance(time, (str, list, np.ndarray)), "time must be a str or array_like"

    out = np.atleast_1d(time).astype("datetime64[ns]")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

from .eis_combine_proton_spec import eis_combine_proton_spec
from .eis_omni import eis_omni

# Local imports
from .eis_pad import eis_pad

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _despin(inp, spin_nums):
    spin_starts = np.where(spin_nums[1:] > spin_nums[:-1])[0]
    time_rec = inp.time.data[spin_starts]

    pad_ds = np.zeros([len(spin_starts), len(inp.theta), len(inp.energy)])

    c_strt = 0

    for i, spin_strt in enumerate(spin_starts):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            pad_ds[i, :, :] = np.nanmean(
                inp.data[c_strt : spin_strt + 1, :, :],
                axis=0,
            )
        c_strt = spin_strt + 1

    out = xr.DataArray(
        pad_ds,
        coords=[time_rec, inp.theta.data, inp.energy.data],
        dims=["time", "theta", "energy"],
    )

    return out


def eis_combine_proton_pad(
    phxtof_allt,
    extof_allt,
    vec: xr.DataArray = None,
    energy: list = None,
    pa_width: int = 15,
    despin: bool = False,
):
    r"""Combines EPD-EIS Energy by Time Of Flight (ExTOF) and Pulse Height by
    Time Of Flight (PHxTOF) proton Pitch-Angle Distributions (PADs).

    Parameters
    ----------
    phxtof_allt : xarray.Dataset
        Dataset containing the energy spectrum of the 6 telescopes of the
        Energy Ion Spectrometer (EIS) for MCP Pulse Height by Time Of Flight
        (PHxTOF).
    extof_allt : xarray.Dataset
        Dataset containing the energy spectrum of the 6 telescopes of the
        Energy Ion Spectrometer (EIS) for Energy by Time Of Flight (ExTOF).
    vec : xarray.DataArray, Optional
        Axis with to compute pitch-angle. Default is X-GSE.
    energy : array_like, Optional
        Energy range to include in the calculation. Default is [55, 800].
    pa_width : int, Optional
        Size of the pitch angle bins, in degrees. Default is 15.
    despin : bool, Optional
        Remove spintone. Default is False.

    Returns
    -------
    out : xarray.DataArray
        Combined PHxTOF and ExTOF proton PADs.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_pad, pyrfu.mms.eis_omni,
    pyrfu.mms.eis_combine_proton_spec


    """

    if energy is None:
        energy = [phxtof_allt.energy.data[0], extof_allt.energy.data[-1]]

    # set up the number of pa bins to create
    n_pabins = int(180.0 / pa_width)
    pa_label = 180.0 * np.arange(n_pabins) / n_pabins + pa_width / 2.0

    # Compute Pulse-Height x Time Of Flight (PHxTOF) Pitch-Angle
    # Distribution (PAD)
    phxtof_pad = eis_pad(phxtof_allt, vec, energy, pa_width)

    # Compute Energy x Time Of Flight (ExTOF) Pitch-Angle Distribution (PAD)
    extof_pad = eis_pad(extof_allt, vec, energy, pa_width)

    # Compute combined PHxTOF and ExTOF omni-directional energy spectrum.
    proton_combined_spec = eis_omni(
        eis_combine_proton_spec(phxtof_allt, extof_allt),
    )

    data_size = [len(phxtof_pad), len(extof_pad)]

    if data_size[0] > data_size[1]:
        time_data = extof_pad.time.data
        phxtof_pad_data = phxtof_pad.data[: data_size[1], ...]
        extof_pad_data = extof_pad.data
    elif data_size[0] < data_size[1]:
        time_data = phxtof_pad.time.data
        phxtof_pad_data = phxtof_pad.data
        extof_pad_data = extof_pad.data[: data_size[0], ...]
    else:
        time_data = phxtof_pad.time.data
        phxtof_pad_data = phxtof_pad.data
        extof_pad_data = extof_pad.data

    cond_ = np.logical_and(
        proton_combined_spec.energy.data > energy[0],
        proton_combined_spec.energy.data < energy[1],
    )
    energy_size = np.where(cond_)[0]

    proton_pad = np.zeros((len(time_data), n_pabins, len(energy_size)))

    cond_phxtof = np.logical_and(
        phxtof_pad.energy.data > 14,
        phxtof_pad.energy.data < 52,
        phxtof_pad.energy.data > energy[0],
    )
    cond_extof = np.logical_and(
        extof_pad.energy.data > 82,
        extof_pad.energy.data < energy[1],
    )

    phxtof_taren = np.where(cond_phxtof)[0]
    phxtof_taren_cross = np.where(phxtof_pad.energy.data > 52)[0]
    extof_taren_cross = np.where(extof_pad.energy.data < 82)[0]
    extof_taren = np.where(cond_extof)[0]

    proton_pad[..., : phxtof_taren.size] = phxtof_pad_data[..., phxtof_taren]

    for (i, phxtof_en), extof_en in zip(
        enumerate(phxtof_taren_cross),
        extof_taren_cross,
    ):
        temp_ = np.stack(
            [phxtof_pad_data[..., phxtof_en], extof_pad_data[..., extof_en]],
        )
        idx_l, idx_r = [phxtof_taren.size, phxtof_taren.size + i + 1]
        proton_pad[..., idx_l:idx_r] = np.nanmean(temp_, axis=0)[:, :, None]

    idx_l = len(phxtof_pad.energy.data)
    idx_r = len(phxtof_pad.energy.data) + len(extof_taren)
    proton_pad[..., idx_l:idx_r] = extof_pad_data[..., extof_taren]

    energy = proton_combined_spec.energy.data[cond_]

    out = xr.DataArray(
        proton_pad,
        coords=[time_data, pa_label, energy],
        dims=["time", "theta", "energy"],
    )

    if despin:
        out = _despin(out, phxtof_allt.spin.data)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def movmean(inp, n_pts: int = 100):
    r"""Computes running average of the inp over npts points.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.
    n_pts : int, Optional
        Number of points to average over.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input variable averaged over npts points.

    Notes
    -----
    Works also with 3D skymap distribution.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load ion pressure tensor

    >>> p_xyz_i = mms.get_data("Pi_gse_fpi_brst_l2", tint, mms_id)

    Running average the pressure tensor over 10s

    >>> fs = pyrf.calc_fs(p_xyz_i)
    >>> p_xyz_i = pyrf.movmean(p_xyz_i, int(10 * fs))

    """

    if isinstance(n_pts, float):
        n_pts = np.floor(n_pts).astype(np.int64)

    if n_pts % 2:
        n_pts -= 1

    # Computes moving average
    cum_sum = np.cumsum(inp.data, axis=0)
    out_dat = (cum_sum[n_pts:, ...] - cum_sum[:-n_pts, ...]) / n_pts

    coords = []

    for k in inp.dims:
        if k == "time":
            coords.append(inp.coords[k][int(n_pts / 2) : -int(n_pts / 2)])
        else:
            coords.append(inp.coords[k].data)

    # Output in DataArray type
    out = xr.DataArray(out_dat, coords=coords, dims=inp.dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np
import xarray as xr

from .datetime642unix import datetime642unix

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _spec_mat(half_spec_x, half_spec_y, half_spec_z):
    r"""CALCULATION OF THE SPECTRAL MATRIX"""

    spec_mat = np.zeros([len(half_spec_x), 3, 3])
    spec_mat[:, 0, 0] = half_spec_x * np.conj(half_spec_x)
    spec_mat[:, 1, 0] = half_spec_x * np.conj(half_spec_y)
    spec_mat[:, 2, 0] = half_spec_x * np.conj(half_spec_z)
    spec_mat[:, 0, 1] = half_spec_y * np.conj(half_spec_x)
    spec_mat[:, 1, 1] = half_spec_y * np.conj(half_spec_y)
    spec_mat[:, 2, 1] = half_spec_y * np.conj(half_spec_z)
    spec_mat[:, 0, 2] = half_spec_z * np.conj(half_spec_x)
    spec_mat[:, 1, 2] = half_spec_z * np.conj(half_spec_y)
    spec_mat[:, 2, 2] = half_spec_z * np.conj(half_spec_z)

    return spec_mat


def wavepolarize_means(
    b_wave,
    b_bgd,
    min_psd: float = 1e-25,
    nop_fft: int = 256,
):
    r"""Analysis the polarization of magnetic wave using "means" method

    Parameters
    -----------
    b_wave : xarray.DataArray
        Time series of the magnetic field from Search Coil Magnetometer (SCM).
    b_bgd : xarray.DataArray
        Time series of the magnetic field from Flux Gate Magnetometer (FGM).
    min_psd : float, Optional
        Threshold for the analysis (e.g 1.0e-7). Below this value, the SVD
        analysis is meaningless if min_psd is not given, SVD analysis will
        be done for all waves. Default ``min_psd`` = 1e-25.
    nop_fft : int, Optional
        Number of points in FFT. Default is 256.

    Returns
    -------
    b_psd : xarray.DataArray
        Power spectrum density of magnetic filed wave.
    wave_angle : xarray.DataArray
        (form 0 to 90)
    deg_pol : xarray.DataArray
        Spectrogram of the degree of polarization (form 0 to 1).
    elliptict : xarray.DataArray
        Spectrogram of the ellipticity (form -1 to 1)
    helict : xarray.DataArray
        Spectrogram of the helicity (form -1 to 1)

    Notes
    ------
    ``b_wave`` and ``b_bgd`` should be from the same satellite and in the same
    coordinates

    .. warning::
        If one component is an order of magnitude or more  greater than the
        other two then the polarization results saturate and erroneously
        indicate high degrees of polarization at all times and frequencies.
        Time series should be eyeballed before running the program. For time
        series containing very rapid changes or spikes the usual problems
        with Fourier analysis arise. Care should be taken in evaluating
        degree of polarization results. For meaningful results there should
        be significant wave power at the frequency where the polarization
        approaches 100%. Remember comparing two straight lines yields 100%
        polarization.

    Examples
    --------
    >>> from pyrfu import pyrf
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd)
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd, 1.0e-7)
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd, 1.0e-7, 256)

    """

    step_length = int(nop_fft / 2)
    n_pts = len(b_wave)
    # total number of FFTs
    n_stp = int((n_pts - nop_fft) / step_length)
    # No. of bins in frequency domain
    n_bin = 7
    # Smoothing profile based on Hanning
    aa = np.array([0.024, 0.093, 0.232, 0.301, 0.232, 0.093, 0.024])

    # change wave to MFA coordinates
    b_bgd = resample(b_bgd, b_wave)

    b_x, b_y, b_z = [np.zeros(len(b_wave)) for _ in range(3)]

    for ii in range(len(b_wave)):
        nb = b_bgd[ii, :] / np.linalg.norm(b_bgd[ii, :])
        n_perp1 = np.cross(nb, [0, 1, 0])
        n_perp1 = n_perp1 / np.linalg.norm(n_perp1)
        n_perp2 = np.cross(nb, n_perp1)

        b_z[ii] = np.sum(b_wave[ii, :] * nb)
        b_x[ii] = np.sum(b_wave[ii, :] * n_perp1)
        b_y[ii] = np.sum(b_wave[ii, :] * n_perp2)

    ct = datetime642unix(b_wave.time.data)

    # DEFINE ARRAYS
    xs, ys, zs = [b_x, b_y, b_z]

    sample_freq = 1 / (ct[1] - ct[0])
    end_sample_freq = 1 / (ct[-1] - ct[-2])

    if sample_freq != end_sample_freq:
        logging.info(
            "file sampling frequency changes %(sample_freq)3.2f Hz "
            "to %(end_sample_freq)3.2f Hz",
            {"sample_freq": sample_freq, "end_sample_freq": end_sample_freq},
        )
    else:
        logging.info("ac file sampling frequency %3.2f Hz", sample_freq)

    # FFT calculation
    # Minimum variance direction and wave normal angle
    wave_angle = np.zeros([n_stp, int(nop_fft / 2)])

    # Degree of Polarization
    sqrd_mat = np.zeros([n_stp, int(nop_fft / 2), 3, 3])
    deg_pol = np.zeros([n_stp, int(nop_fft / 2)])
    trace_spec_mat = np.zeros([n_stp, int(nop_fft / 2)])

    # HELICITY, ELLIPTICITY AND THE WAVE STATE VECTOR
    lambda_u = np.zeros([n_stp, int(nop_fft / 2), 3, 3])

    helic, ellip = [np.zeros([n_stp, int(nop_fft / 2), 3]) for _ in range(2)]

    smooth = np.zeros(nop_fft)

    for j in range(n_stp):
        # FFT CALCULATION
        smooth = 0.08 + 0.46 * (
            1 - np.cos(2 * np.pi * np.arange(1, nop_fft + 1) / nop_fft)
        )
        temp_x = (
            smooth * xs[((j - 1) * step_length + 1) : ((j - 1) * step_length + nop_fft)]
        )
        temp_y = (
            smooth * ys[((j - 1) * step_length + 1) : ((j - 1) * step_length + nop_fft)]
        )
        temp_z = (
            smooth * zs[((j - 1) * step_length + 1) : ((j - 1) * step_length + nop_fft)]
        )

        spec_x = np.fft.fft(temp_x)
        spec_y = np.fft.fft(temp_y)
        spec_z = np.fft.fft(temp_z)

        half_spec_x = spec_x[: (nop_fft / 2)]
        half_spec_y = spec_y[: (nop_fft / 2)]
        half_spec_z = spec_z[: (nop_fft / 2)]

        xs = np.roll(xs, -step_length)
        ys = np.roll(ys, -step_length)
        zs = np.roll(zs, -step_length)

        # CALCULATION OF THE SPECTRAL MATRIX
        spec_mat = _spec_mat(half_spec_x, half_spec_y, half_spec_z)

        # Calculation of smoothed spectral matrix
        e_spec_mat = np.nan * np.ones(spec_mat.shape)

        off_idx = int((n_bin - 1) / 2)

        for k in range(off_idx, int(nop_fft / 2) - off_idx):
            for ir in range(3):
                for ic in range(3):
                    e_spec_mat[k, ir, ic] = np.sum(
                        aa[:n_bin] * spec_mat[(k - off_idx) : (k + off_idx), ir, ic],
                    )

        # Calculation of the minimum variance direction and wave normal angle
        aaa2 = np.imag(e_spec_mat[:, 0, 1]) ** 2
        aaa2 += np.imag(e_spec_mat[:, 0, 2]) ** 2
        aaa2 += np.imag(e_spec_mat[:, 1, 2]) ** 2
        aaa2 = np.sqrt(aaa2[j, :])

        wn_x = -np.abs(np.imag(e_spec_mat[:, 1, 2]) / aaa2)
        wn_y = -np.abs(np.imag(e_spec_mat[:, 0, 2]) / aaa2)
        wn_z = np.imag(e_spec_mat[:, 0, 1]) / aaa2

        wave_angle[j, :] = np.arctan(
            np.sqrt(wn_x**2 + wn_y**2) / np.abs(wn_z),
        )

        # CALCULATION OF THE DEGREE OF POLARISATION
        # calc of square of smoothed spec matrix
        for ir in range(3):
            for ic in range(3):
                sqrd_mat[:, ir, ic] = e_spec_mat[:, ir, 0] * e_spec_mat[:, 0, ic]
                sqrd_mat[:, ir, ic] += e_spec_mat[:, ir, 1] * e_spec_mat[:, 1, ic]
                sqrd_mat[:, ir, ic] += e_spec_mat[:, ir, 2] * e_spec_mat[:, 2, ic]

        trace_sqrd_mat = sqrd_mat[:, 0, 0] + sqrd_mat[:, 1, 1] + sqrd_mat[:, 2, 2]
        trace_spec_mat[j, :] = (
            e_spec_mat[:, 0, 0] + e_spec_mat[:, 1, 1] + e_spec_mat[:, 2, 2]
        )

        deg_pol[j, :] = trace_spec_mat[j, :] * np.nan
        deg_pol[j, off_idx : int(nop_fft / 2) - off_idx] = (
            3 * trace_sqrd_mat[off_idx : int(nop_fft / 2) - off_idx]
        )
        deg_pol[j, off_idx : int(nop_fft / 2) - off_idx] -= (
            trace_spec_mat[j, off_idx : int(nop_fft / 2) - off_idx] ** 2
        )
        deg_pol[j, off_idx : int(nop_fft / 2) - off_idx] /= (
            2 * trace_spec_mat[j, off_idx : int(nop_fft / 2) - off_idx] ** 2
        )

        # Calculation of helicity, ellipticity and the wave state vector
        alpha_x = np.sqrt(e_spec_mat[:, 0, 0])
        alpha_y = np.sqrt(e_spec_mat[:, 1, 1])
        alpha_z = np.sqrt(e_spec_mat[:, 2, 2])

        alpha_cos1_x = np.real(e_spec_mat[:, 0, 1]) / np.sqrt(
            e_spec_mat[:, 0, 0],
        )
        alpha_sin1_x = -np.imag(e_spec_mat[j, :, 0, 1]) / np.sqrt(
            e_spec_mat[:, 0, 0],
        )
        alpha_cos2_x = np.real(e_spec_mat[:, 0, 2]) / np.sqrt(
            e_spec_mat[:, 0, 0],
        )
        alpha_sin2_x = -np.imag(e_spec_mat[j, :, 0, 2]) / np.sqrt(
            e_spec_mat[:, 0, 0],
        )

        alpha_cos1_y = np.real(e_spec_mat[:, 1, 0]) / np.sqrt(
            e_spec_mat[:, 1, 1],
        )
        alpha_sin1_y = -np.imag(e_spec_mat[:, 1, 0]) / np.sqrt(
            e_spec_mat[:, 1, 1],
        )
        alpha_cos2_y = np.real(e_spec_mat[:, 1, 2]) / np.sqrt(
            e_spec_mat[:, 1, 1],
        )
        alpha_sin2_y = -np.imag(e_spec_mat[:, 1, 2]) / np.sqrt(
            e_spec_mat[:, 1, 1],
        )

        alpha_cos1_z = np.real(e_spec_mat[:, 2, 0]) / np.sqrt(
            e_spec_mat[:, 2, 2],
        )
        alpha_sin1_z = -np.imag(e_spec_mat[:, 2, 0]) / np.sqrt(
            e_spec_mat[:, 2, 2],
        )
        alpha_cos2_z = np.real(e_spec_mat[:, 2, 1]) / np.sqrt(
            e_spec_mat[:, 2, 2],
        )
        alpha_sin2_z = -np.imag(e_spec_mat[:, 2, 1]) / np.sqrt(
            e_spec_mat[:, 2, 2],
        )

        lambda_u[:, 0, 0] = alpha_x
        lambda_u[:, 1, 0] = alpha_y
        lambda_u[:, 2, 0] = alpha_z

        lambda_u[:, 0, 1] = alpha_cos1_x + alpha_sin1_x * 1j
        lambda_u[:, 0, 2] = alpha_cos2_x + alpha_sin2_x * 1j

        lambda_u[:, 1, 1] = alpha_cos1_y + alpha_sin1_y * 1j
        lambda_u[:, 1, 2] = alpha_cos2_y + alpha_sin2_y * 1j

        lambda_u[:, 2, 1] = alpha_cos1_z + alpha_sin1_z * 1j
        lambda_u[:, 2, 2] = alpha_cos2_z + alpha_sin2_z * 1j

        for k in range(int(nop_fft / 2)):
            for xyz in range(3):
                # HELICITY CALCULATION
                upper = np.sum(
                    2 * np.real(lambda_u[k, xyz, :3]) * (np.imag(lambda_u[k, xyz, :3])),
                )
                lower = np.sum(
                    (np.real(lambda_u[k, xyz, :3])) ** 2
                    - (np.imag(lambda_u[k, xyz, :3])) ** 2,
                )

                if upper[j, k] > 0:
                    gamma = np.arctan(upper[j, k] / lower[j, k])
                else:
                    gamma = np.pi + (np.pi + np.arctan(upper[j, k] / lower[j, k]))

                lambda_u[k, xyz, :] = np.exp(-0.5 * gamma * 1j) * lambda_u[k, xyz, :]
                helic[j, k, xyz] = np.sqrt(
                    np.sum(np.real(lambda_u[k, xyz, :3]) ** 2),
                )
                helic[j, k, xyz] /= np.sqrt(
                    np.sum(np.imag(lambda_u[k, xyz, :3]) ** 2),
                )
                helic[j, k, xyz] = np.divide(1, helic[j, k, xyz])

                # ELLIPTICITY CALCULATION
                upper_e = np.sum(
                    np.imag(lambda_u[k, xyz, :3]) * np.real(lambda_u[k, xyz, :3]),
                )
                lower_e = np.sum(np.real(lambda_u[k, xyz, :2]) ** 2) - np.sum(
                    np.imag(lambda_u[k, xyz, :2]) ** 2,
                )

                if upper_e > 0:
                    gamma_rot = np.arctan(upper_e / lower_e)
                else:
                    gamma_rot = np.pi + np.pi + np.arctan(upper_e / lower_e)

                lam = lambda_u[k, xyz, :2]
                lambda_u_rot = np.exp(-0.5 * gamma_rot * 1j) * lam

                ellip[j, k, xyz] = np.sqrt(np.sum(np.imag(lambda_u_rot) ** 2))
                ellip[j, k, xyz] /= np.sqrt(
                    np.sum(np.real(lambda_u_rot) ** 2),
                )
                ellip[j, k, xyz] *= -(
                    np.imag(e_spec_mat[k, 0, 1]) * np.sin(wave_angle[j, k])
                )
                ellip[j, k, xyz] /= np.abs(
                    np.imag(e_spec_mat[k, 0, 1]) * np.sin(wave_angle[j, k]),
                )

    # AVERAGING HELICITY AND ELLIPTICITY RESULTS
    ellipticity = np.mean(ellip, axis=-1)
    helicity = np.mean(helic, axis=-1)

    # CREATING OUTPUT PARAMETER
    time_line = (
        ct[0]
        + np.abs(nop_fft / 2) / sample_freq
        + np.arange(1, n_stp + 1) * step_length / sample_freq
    )
    bin_width = sample_freq / nop_fft
    freq_line = bin_width * np.arange(1, nop_fft / 2 + 1)

    # scaling power results to units with meaning
    W = nop_fft * np.sum(smooth**2)

    power_spec = np.zeros([n_stp, int(nop_fft / 2)])
    power_spec[:, 1 : nop_fft / 2 - 1] = (
        1 / W * 2 * trace_spec_mat[:, 1 : nop_fft / 2 - 1] / bin_width
    )
    power_spec[:, 1] = 1 / W * trace_spec_mat[:, 0] / bin_width
    power_spec[:, nop_fft / 2] = 1 / W * trace_spec_mat[:, nop_fft / 2] / bin_width

    # KICK OUT THE ANALYSIS OF THE WEAK SIGNALS
    wave_angle[power_spec < min_psd] = np.nan
    deg_pol[power_spec < min_psd] = np.nan
    ellipticity[power_spec < min_psd] = np.nan
    helicity[power_spec < min_psd] = np.nan

    # Save as DataArrays
    b_psd = xr.DataArray(
        power_spec,
        coords=[time_line, freq_line],
        dims=["t", "f"],
    )
    wave_angle = xr.DataArray(
        wave_angle * 180 / np.pi,
        coords=[time_line, freq_line],
        dims=["t", "f"],
    )
    ellipticity = xr.DataArray(
        ellipticity,
        coords=[time_line, freq_line],
        dims=["t", "f"],
    )
    deg_pol = xr.DataArray(
        deg_pol,
        coords=[time_line, freq_line],
        dims=["t", "f"],
    )
    helicity = xr.DataArray(
        helicity,
        coords=[time_line, freq_line],
        dims=["t", "f"],
    )

    b_psd.f.attrs["units"] = "Hz"
    wave_angle.f.attrs["units"] = "Hz"
    deg_pol.f.attrs["units"] = "Hz"
    ellipticity.f.attrs["units"] = "Hz"
    helicity.f.attrs["units"] = "Hz"

    return b_psd, wave_angle, deg_pol, ellipticity, helicity
